#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "m.h"
#include "btree.h"
#include "g.h"
#include "fit.h"

static int fitbtree_npar( struct SBTREE *);
static char rcsid[] = {"$Id: fitbtree.c,v 1.1.1.1 1997/07/10 17:59:57 kraemer Exp $"};

/*
   --- Binary tree evaluation function

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitbtree.c,v $
   Revision 1.1.1.1  1997/07/10 17:59:57  kraemer
   revision for cvs


*/
#define DDPAR(ipar,ix) ddPar[ (ipar) * nX + (ix) ]
int fitbtree(
              struct SFITFCT *psFF,   /* ptr. to descr. of 1 function */
              double *dPar,           /* function parameters */
              double *ddPar,          /* unused sofar */
              double *dX, double *dY,
              int nPar, int nX,
              unsigned long lFlag ) {

   int ii, iRC = ERRFIT_SYS;

   struct SBTREE *psBT = psFF -> pvBT;
   struct SBTREEVIDF *psBTI=NULL, *psBTIalloc = NULL;

   switch( lFlag ) {
   case FITFCT_NPAR:                  /* return # of parameters */
      if ( (nPar = fitbtree_npar(psBT)) <= 0 ) return(nPar);
                                      /* build identifier table
                                         "x", "p1"..."pn" */
      if ( ( psBTI = psBTIalloc =
             calloc( nPar+2, sizeof(struct SBTREEVIDF) ) ) == NULL ) {
         iRC = FITUSR_SYS; goto gError;
      }
      if ( (psBTI -> cIdf = strdup("x")) == NULL ) {
         iRC = FITUSR_SYS; goto gError;
      }
      psBTI++;
      for ( ii=0; ii<nPar; ii++ ) {
         if ( (psBTI -> cIdf = calloc(1,10) ) == NULL ) {
            iRC = FITUSR_SYS; goto gError;
         }
         sprintf( psBTI -> cIdf, "p%d", ii+1 );
         psBTI++;
      }
      if ( ( iRC = btreeprc( psBT, psBTIalloc, 0, BTREE_RESOLVEVIDF ) ) != 0 ){
         iRC = FITUSR_IDFAILED; break;
      }
      psFF -> pvBTI = psBTIalloc; psBTIalloc = NULL;
      iRC = nPar;
   break;
   case FITFCT_VAL:                   /* return start values    */
      vfill ( 0E0, dPar, 1, nPar ); 
      iRC = FITUSR_OK;
   break;
   case FITFCT_FCT:                   /* return function values only */
   case FITFCT_DISP:                  /* return function values only, call for display */
   case FITFCT_RESULT:                /* return final function values */
      ddPar = NULL;
   case FITFCT_DERIV:                 /* return function values & parameter derivatives  */
                                      /* complete the ID table */
#if defined(SCALAR)
      psBTI = psFF -> pvBTI; psBTI++; /* skip "x" entry */
#endif
      psBTI = psFF -> pvBTI;
      psBTI -> pIdf = dX; psBTI -> ii = 1; psBTI -> nn = nX;
                                      /* loop to insert actual parameter addresses */
      for ( ii=0;ii<nPar; ii++ ) {
         psBTI++;
         psBTI -> pIdf = dPar+ii; /*psBTI -> ii = 0; psBTI -> nn = 0;*/
      }
      iRC = btreeprc( psBT, psFF->pvBTI, 0, BTREE_RESOLVEVIDF );
      if ( iRC != 0 ) {
         iRC = FITUSR_IDFAILED; goto gError;
      }
      iRC = btreeprc( psBT, NULL, 0, BTREE_EVALUATE );
      if ( iRC != 0 ) {
         iRC = FITUSR_FCTFAILED; goto gError;
      }
      vmov( psBT -> sBTD[0].pd, psBT -> sBTD[0].sDim[0].lStride,
            dY, 1, nX );
#if defined(SCALAR)
                                      /* insert x-value addresses */
      psBTI = psFF -> pvBTI;
      for ( ii=0; ii<nX; ii++ ) {
         psBTI -> pIdf = dX+ii;
         iRC = btreeprc( psBT, psFF->pvBTI, 0, BTREE_RESOLVEVIDF );
         if ( iRC != 0 ) return(ERRFIT_FCTEVAL);
         iRC = btreeprc( psBT, NULL, 0, BTREE_EVALUATE );
         if ( iRC != 0 ) return(ERRFIT_FCTEVAL);
         dY[ii] = psBT->sBTD[0].pd[0];
      }
#endif
      iRC = FITUSR_NODERIV;           /* no derivatives returned */
   break;
   case FITFCT_CLEANUP:
      if ( psBTI = psFF -> pvBTI ) {
         for ( ii=0; ii<psFF->np+1; ii++ ) {
            if (psBTI->cIdf) free(psBTI->cIdf); psBTI++;
         }
         free( psFF -> pvBTI );
      }
      iRC = FITUSR_OK;
   break;
   default: iRC = FITUSR_OK; break;
   }
gError:;
                                      /* cleanup ID table */
   if ( psBTI = psBTIalloc ) {
      for ( ii=0; ii<psFF->np+1; ii++ ) {
         if (psBTI->cIdf) free(psBTI->cIdf); psBTI++;
      }
      free( psBTIalloc );
   }
   return(iRC);
}
/*
   --- Auxiliary function to search recursively for "p#"
       to determine max. # of parameters
*/ 
static int fitbtree_npar( struct SBTREE *psBT ) {
   int nPar=0, iRC, ii;

   if ( psBT -> cOp ) {
      for ( ii=0; ii<2; ii++ ) {
         if ( psBT->vSon[ii] ) {
            if ( ( iRC = fitbtree_npar( psBT->vSon[ii] ) ) < 0 ) return(iRC);
            nPar = max( nPar, iRC );
         }
      }
   } else {
      if ( ( iRC = sscanf( psBT -> vSon[0], "p%d", &ii ) ) == 1 )
         nPar = max( nPar, ii );
      if ( ( iRC = sscanf( psBT -> vSon[0], "P%d", &ii ) ) == 1 )
         nPar = max( nPar, ii );
   }
   return(nPar);
}
