
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <limits.h>

#include "u.h"
#include "m.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitcmdleg.c,v 1.1.1.1 1997/07/10 17:59:59 kraemer Exp $"};

extern struct SFIT sFit;
/*
   --- Command processor for the FLEGENDRE command

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitcmdleg.c,v $
   Revision 1.1.1.1  1997/07/10 17:59:59  kraemer
   revision for cvs

*/
                                      /* local flag bits */
#define fitCmdLegEVEN 0x00000001
#define fitCmdLegODD  0x00000002
#define fitCmdLegNORM 0x00000004

void fitCmdLeg(
                long *plOrder,
                long *plDeg, long *plRad,
                long *plEven, long *plOdd,
                long *plNorm,
                long *plDebug
              ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__, *pMsg = pDbgMsg;
   unsigned long lFlag = 0;

   unsigned long lType = 0, lStatus = 0;
   int np = 0;
   struct SFITFCT sFF;
   char cStr[255+1];


   if ( *plDebug ) lFlag |= FIT_DEBUG;
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sFF, 0, sizeof(sFF) );

                                      /* check inputs */
   if ( *plOrder < 0 ) {
      iRC = ERRFIT_CMDLEGORDER; goto gError;
   }
                                      /* degree/radians */
   if ( *plRad ) {
      lStatus |= FITFCT_RAD;
   }
   if ( *plDeg ) {
      lStatus |= FITFCT_DEG;
   }
   if ( ( lStatus & (FITFCT_RAD|FITFCT_DEG) ) == (FITFCT_RAD|FITFCT_DEG) ) {
      pMsg = "rad,deg";
      iRC = ERRFIT_CMDPAREXCL; goto gError;
   }
                                      /* set legendre type locally */
   lType = 0;
   if ( *plEven ) lType |= fitCmdLegEVEN;
   if ( *plOdd  ) lType |= fitCmdLegODD;
   if ( *plNorm ) lType |= fitCmdLegNORM;
/*
fprintf( stderr, "<D> %d %X\n", lType, lType );
*/
   switch ( lType ) {
   case fitCmdLegODD:
   case fitCmdLegODD | fitCmdLegNORM:
      np = (*plOrder+1)/2 + 1;
   break;
   case fitCmdLegEVEN:
   case fitCmdLegEVEN | fitCmdLegNORM:
      np = *plOrder/2 + 1;
   break;
   case 0: case ( fitCmdLegEVEN | fitCmdLegODD ):
      np = *plOrder   + 1;
      lType = fitCmdLegEVEN | fitCmdLegODD;
   break;
   case fitCmdLegNORM:
      np = *plOrder   + 1;
      lType = fitCmdLegEVEN | fitCmdLegODD | fitCmdLegNORM;
   break;
   default:
      iRC = ERRFIT_FCTUNKNOWN; goto gError;
   break;
   }
                                      /* allocate local function  */
   if ( ( iRC = fitFctAlloc( np, &sFF, lFlag) ) != 0 ) goto gError;
                                      /* fill entries */
   fitFctInit( &sFF, lFlag);
   sFF.lStatus |= lStatus;
   switch( lType ) {
   case fitCmdLegEVEN:
      sprintf( cStr, "Even Legendre polynomial of order %ld", *plOrder );
      sFF.pcName = strdup("Legendre polynomial");
      sFF.eCall  = fitLegEven;
      sFF.lType  = FITFCT_LEGEVEN;
   break;
   case fitCmdLegODD:
      sprintf( cStr, "Odd  Legendre polynomial of order %ld", *plOrder );
      sFF.pcName = strdup("Legendre polynomial");
      sFF.eCall  = fitLegOdd;
      sFF.lType  = FITFCT_LEGODD;
   break;
   case fitCmdLegEVEN | fitCmdLegODD:
      sprintf( cStr, "Legendre polynomial of order %ld", *plOrder );
      sFF.pcName = strdup("Legendre polynomial");
      sFF.eCall  = fitLeg;
      sFF.lType  = FITFCT_LEG;
   break;
   case fitCmdLegEVEN | fitCmdLegNORM:
      sprintf( cStr, "Normalized even Legendre polynomial of order %ld", *plOrder );
      sFF.pcName = strdup("Normalized Legendre polynomial");
      sFF.eCall  = fitLegEvenNorm;
      sFF.lType  = FITFCT_LEGEVENNORM;
   break;
   case fitCmdLegODD | fitCmdLegNORM:
      sprintf( cStr, "Normalized odd  Legendre polynomial of order %ld", *plOrder );
      sFF.pcName = strdup("Normalized Legendre polynomial");
      sFF.eCall  = fitLegOddNorm;
      sFF.lType  = FITFCT_LEGODDNORM;
   break;
   case fitCmdLegEVEN | fitCmdLegODD | fitCmdLegNORM:
      sprintf( cStr, "Normalized Legendre polynomial of order %ld", *plOrder );
      sFF.pcName = strdup("Normalized Legendre polynomial");
      sFF.eCall  = fitLegNorm;
      sFF.lType  = FITFCT_LEGNORM;
   break;
   default:
      iRC = ERRFIT_FCTUNKNOWN; goto gError;
   break;
   }
   if ( ( sFF.pcText = strdup(cStr) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }

   sFF.npSpecial = *plOrder;

                                      /* initialize parameters */
   fitDataFctInitPar( &sFit.sFD, &sFF, lFlag );
                                      /* add to function block */
   if ( ( iRC = fitFctBlkFctPut( &sFit.sFFB, &sFF, lFlag) ) != 0 ) goto gError;
                                      /*  */

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
   fitFctFree( &sFF, lFlag);
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return;
}
