#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <limits.h>

#include "m.h"
#include "cmd.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitcmdpar.c,v 1.2 1997/12/15 19:03:46 kraemer Exp $"};

extern struct SFIT sFit;
/*
   --- Command processor for the FPAR command

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitcmdpar.c,v $
   Revision 1.2  1997/12/15 19:03:46  kraemer
   fit statistics part of function block

   Revision 1.1.1.1  1997/07/10 18:00:00  kraemer
   revision for cvs

*/

void fitCmdPar(
                const char *pcPar,
                long *plFix,
                long *plDiff, long *plSum, long *plQuot, long *plProd,
                long *plVar, long *plEOn, long *plEOff,
                const char *pcGuess, long *plRel,
                const char *pcMin, const char *pcMax, long *plNoLim,
                long *plList, long *plSwitch, long *plDebug
              ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__, *pMsg = pDbgMsg;
   unsigned long lFlag = 0, lFlagCurve = 0;

   unsigned long lFlagPeak = 0;
   long *plvPar = NULL, *plvPeak = NULL; int nPar=0;
   double *pdvGuess = NULL; int nGuess=0;
   double *pdvMin = NULL; int nMin=0;
   double *pdvMax = NULL; int nMax=0;

   double dVal, dRefVal;
   long lRefNum = 0, ll;
   int ii, ip, jp;
   struct SFITPAR sFP, sFPRef;


   if ( *plDebug ) lFlag |= FIT_DEBUG;
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   if ( sFit.lStatus & FIT_POISSON ) lFlagCurve |= FITFCTBLK_POISSON;
   memset( &sFPRef, 0, sizeof(sFPRef) );
                                      /* check valid inputs */
   if ( *plFix && *plVar ) {
      pMsg = "fix,var";
      iRC = ERRFIT_CMDPAREXCL; goto gError;
   }
   if ( *plEOn && *plEOff ) {
      pMsg = "eon,eoff";
      iRC = ERRFIT_CMDPAREXCL; goto gError;
   }
                                      /* some specials for area/pos/width*/
   switch(*plSwitch) {
   case FITCMDSWITCH_AREA:
      lFlagPeak |= FITFCTBLKPAR_AREA;  *plProd = 0;
   break;
   case FITCMDSWITCH_POS:
      lFlagPeak |= FITFCTBLKPAR_POS;   *plProd = 0; *plSum = 0;
   break;
   case FITCMDSWITCH_WIDTH:
      lFlagPeak |= FITFCTBLKPAR_WIDTH; *plProd = 0; *plSum = 0; *plDiff = 0;
   break;
   default:
   break;
   }
/*
   --- handle reference parameters
*/
   ii = 0;
   if ( *plSum >0 ) { ii++; lRefNum = *plSum; }
   if ( *plDiff>0 ) { ii++; lRefNum = *plDiff;}
   if ( *plProd>0 ) { ii++; lRefNum = *plProd;}
   if ( *plQuot>0 ) { ii++; lRefNum = *plQuot;}
   if ( ii>1 ) {
      pMsg = "sum(),diff(),prod(),quot()";
      iRC = ERRFIT_CMDPAREXCL; goto gError;
   }

                                      /* get reference parameter */
   if ( lRefNum > 0 ) {
                                      /* special for peak parameters */
      ll = lRefNum;
      switch(*plSwitch) {
      case FITCMDSWITCH_AREA:
         if ( iRC = fitFctBlkPeak2Par(&sFit.sFFB,&ll,1,&lRefNum,lFlag) ) goto gError;
         lRefNum += FITFCT_OFFSPEAKAREA;
      break;
      case FITCMDSWITCH_POS:
         if ( iRC = fitFctBlkPeak2Par(&sFit.sFFB,&ll,1,&lRefNum,lFlag) ) goto gError;
         lRefNum += FITFCT_OFFSPEAKPOS;
      break;
      case FITCMDSWITCH_WIDTH:
         if ( iRC = fitFctBlkPeak2Par(&sFit.sFFB,&ll,1,&lRefNum,lFlag) ) goto gError;
         lRefNum += FITFCT_OFFSPEAKWIDTH;
      break;
      default:
      break;
      }

      if ( (iRC=fitFctBlkParGetByNum( &sFit.sFFB, lRefNum, &sFPRef, lFlag|FIT_NOMSG ))!= 0 ) goto gError;
                                      /* can't be relative itself */
      if ( sFPRef.lRefTyp > 0 ) {
         iRC = ERRFIT_PARREFRELATIVE; goto gError;
      }
   }
/*
   --- parameter/peak numbers
*/
   while( *pcPar == ' ' ) pcPar++;
   switch( *pcPar ) {
                                      /* all peaks/functions */
   case '\0':
   case '*':
      if ( (nPar = sFit.sFFB.np) <= 0 ) {
         iRC = ERRFIT_NOPAR; goto gError;
      }
      if ( (plvPar = calloc( nPar, sizeof(*plvPar) ) ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }

      switch(*plSwitch) {
      case FITCMDSWITCH_AREA:
      case FITCMDSWITCH_POS:
      case FITCMDSWITCH_WIDTH:
                                      /* find all peaks */
         jp = 0;
         for ( ip=0; ip<nPar; ip++ ) {
            ll = ip+1;
            iRC = fitFctBlkPeak2Par(&sFit.sFFB,&ll,1,plvPar+jp,lFlag|FIT_NOMSG);
            if ( iRC == 0 ) jp++;
         }
         if ( (nPar = jp) <= 0 ) {
            iRC = ERRFIT_NOPEAKFUNCTION; goto gError;
         }
      break;
      default:
                                      /* all functions */
         for ( ip=0; ip<nPar; ip++ ) plvPar[ip] = ip+1;
      break;
      }
   break;
                                      /* selected functions */
   default:
                                      /* count 'em */
      if ( (nPar = c2vl( (char *) pcPar, plvPar, 1, INT_MAX )) <= 0 ) {
         pMsg = (char *) pcPar; iRC = ERRFIT_CONVNAN; goto gError;
      }
      if ( (plvPar = calloc( nPar, sizeof(*plvPar) ) ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
                                      /* store 'em */
      switch(*plSwitch) {
      case FITCMDSWITCH_AREA:
      case FITCMDSWITCH_POS:
      case FITCMDSWITCH_WIDTH:
                                      /* peak numbers to parameter numbers*/
         if ( (plvPeak = calloc( nPar, sizeof(*plvPeak) ) ) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         if ( ( c2vl( (char *) pcPar, plvPeak, 1, nPar ) ) != nPar ) {
            pMsg = (char *) pcPar; iRC = ERRFIT_CONVNAN; goto gError;
         }
         if ( iRC = fitFctBlkPeak2Par(&sFit.sFFB,plvPeak,nPar,plvPar,lFlag) )
            goto gError;
      break;
      default:
         if ( ( c2vl( (char *) pcPar, plvPar, 1, nPar ) ) != nPar ) {
            pMsg = (char *) pcPar; iRC = ERRFIT_CONVNAN; goto gError;
         }
      break;
      }

   break;
   }
                                      /* apply offsets for peak par.s */
   switch(*plSwitch) {
   case FITCMDSWITCH_AREA:
      for ( ip=0; ip<nPar; ip++ ) {
         plvPar[ip] += FITFCT_OFFSPEAKAREA;
      }
   break;
   case FITCMDSWITCH_POS:
      for ( ip=0; ip<nPar; ip++ ) {
         plvPar[ip] += FITFCT_OFFSPEAKPOS;
      }
   break;
   case FITCMDSWITCH_WIDTH:
      for ( ip=0; ip<nPar; ip++ ) {
         plvPar[ip] += FITFCT_OFFSPEAKWIDTH;
      }
   break;
   default:
   break;
   }
/*
   --- guess values for parameters
*/
   while( *pcGuess == ' ' ) pcGuess++;
   if ( *pcGuess ) {
                                      /* count 'em */
      if ( (nGuess = c2vd( (char *) pcGuess, pdvGuess, 1, INT_MAX )) <= 0 ) {
         pMsg = (char *) pcGuess; iRC = ERRFIT_CONVNAN; goto gError;
      }
      if ( (nGuess > 1) && (nGuess != nPar) ) {
         pMsg = (char *) pcGuess; iRC = ERRFIT_VALCOUNTMATCHPAR; goto gError;
      }
                                      /* allocate/initialize */
      if ( (pdvGuess = calloc( nGuess, sizeof(*pdvGuess) ) ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
      vfill( DBL_MAX, pdvGuess, 1, nGuess );
                                      /* store 'em */
      if ( ( c2vd( (char *) pcGuess, pdvGuess, 1, nGuess ) ) != nGuess ) {
         pMsg = (char *) pcGuess; iRC = ERRFIT_CONVNAN; goto gError;
      }
   }
/*
   --- values for lower limits
*/
   while( *pcMin == ' ' ) pcMin++;
   if ( *pcMin ) {
                                      /* count 'em */
      if ( (nMin = c2vd( (char *) pcMin, pdvMin, 1, INT_MAX )) <= 0 ) {
         pMsg = (char *) pcMin; iRC = ERRFIT_CONVNAN; goto gError;
      }
      if ( (nMin > 1) && (nMin != nPar) ) {
         pMsg = (char *) pcMin; iRC = ERRFIT_VALCOUNTMATCHPAR; goto gError;
      }
                                      /* allocate/initialize */
      if ( (pdvMin = calloc( nMin, sizeof(*pdvMin) ) ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
      vfill( DBL_MAX, pdvMin, 1, nMin );
                                      /* store 'em */
      if ( ( c2vd( (char *) pcMin, pdvMin, 1, nMin ) ) != nMin ) {
         pMsg = (char *) pcMin; iRC = ERRFIT_CONVNAN; goto gError;
      }
   }
/*
   --- values for upper limits
*/
   while( *pcMax == ' ' ) pcMax++;
   if ( *pcMax ) {
                                      /* count 'em */
      if ( (nMax = c2vd( (char *) pcMax, pdvMax, 1, INT_MAX )) <= 0 ) {
         pMsg = (char *) pcMax; iRC = ERRFIT_CONVNAN; goto gError;
      }
      if ( (nMax > 1) && (nMax != nPar) ) {
         pMsg = (char *) pcMax; iRC = ERRFIT_VALCOUNTMATCHPAR; goto gError;
      }
                                      /* allocate/initialize */
      if ( (pdvMax = calloc( nMax, sizeof(*pdvMax) ) ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
      vfill( DBL_MAX, pdvMax, 1, nMax );
                                      /* store 'em */
      if ( ( c2vd( (char *) pcMax, pdvMax, 1, nMax ) ) != nMax ) {
         pMsg = (char *) pcMax; iRC = ERRFIT_CONVNAN; goto gError;
      }
   }
/*
   --- assign values
*/
   for ( ip=0; ip<nPar; ip++ ) {
      memset( &sFP, 0, sizeof(sFP) );
                                      /* get stored parameter values */
      if ( (iRC=fitFctBlkParGetByNum( &sFit.sFFB, plvPar[ip], &sFP, lFlag|FIT_NOMSG ))!= 0 ) goto gError;
                                      /* check reference parameter */
      if ( lRefNum > 0 ) {
                                      /* can't be itself */
         if ( lRefNum == plvPar[ip] ) {
            iRC = ERRFIT_PARSELFRELATIVE; goto gError;
         }
      }
                                      /* apply flags */
      if ( *plFix ) sFP.lFlag |= FITPAR_FIX;
      if ( *plVar ) {
                    sFP.lFlag &=~FITPAR_FIX;
         sFP.lRefTyp = 0;
         sFP.lRefNum = 0;
         sFP.dRefVal = 0E0;
      }
      if ( *plEOn ) sFP.lFlag |= FITPAR_EON;
      if ( *plEOff) sFP.lFlag &=~FITPAR_EON;
                                      /* apply guess value */
      ii = ip;
      switch( nGuess ) {
      case 0: break;
      case 1:
         ii = 0;                      /* one value for all */
      default:
         if ( ( dVal = dRefVal = pdvGuess[ii] ) != DBL_MAX ) {
                                      /* relatively fixed parameter */
            if ( lRefNum > 0 ) {
               if ( *plRel ) {        /* relative */
                  if (   *plDiff > 0   ) dVal = dRefVal + sFPRef.dVal;
                  if (   *plSum  > 0   ) dVal = dRefVal - sFPRef.dVal;
                  if (   *plQuot > 0   ) dVal = dRefVal * sFPRef.dVal;
                  if ( ( *plProd > 0 ) && ( sFPRef.dVal != 0E0 ) ) dVal = dRefVal / sFPRef.dVal;
               } else {               /* absolute */
                  if (   *plDiff > 0   ) dRefVal -= sFPRef.dVal;
                  if (   *plSum  > 0   ) dRefVal += sFPRef.dVal;
                  if (   *plProd > 0   ) dRefVal *= sFPRef.dVal;
                  if ( ( *plQuot > 0 ) && ( sFPRef.dVal != 0E0 ) ) dRefVal /= sFPRef.dVal;
               }
            }
                                      /* */
            if ( ( dVal >= sFP.dLim[0] ) &&
                 ( dVal <= sFP.dLim[1] ) ) {
               sFP.dVal = dVal;
                                      /* remember reference type and number */
               if ( lRefNum > 0 ) {
                  if ( *plDiff > 0 ) sFP.lRefTyp = FITPAR_DIFF;
                  if ( *plSum  > 0 ) sFP.lRefTyp = FITPAR_SUM;
                  if ( *plProd > 0 ) sFP.lRefTyp = FITPAR_PROD;
                  if ( *plQuot > 0 ) sFP.lRefTyp = FITPAR_QUOT;
                  sFP.lRefNum = lRefNum-1;
                  sFP.dRefVal = dRefVal;
                  sFP.lFlag  |= FITPAR_FIX;
               }
            }
         }
      break;
      }
                                      /* apply lower limit */
      ii = ip;
      switch( nMin ) {
      case 0:
         if ( *plNoLim ) sFP.dLim[0] = -DBL_MAX;
      break;
      case 1:
         ii = 0;                      /* one value for all */
      default:
         if ( pdvMin[ii] != DBL_MAX ) {
            sFP.dLim[0] = min( sFP.dVal, pdvMin[ii] );
         } else {
            if ( *plNoLim ) sFP.dLim[0] = -DBL_MAX;
         }
      break;
      }
                                      /* apply upper limit */
      ii = ip;
      switch( nMax ) {
      case 0:
         if ( *plNoLim ) sFP.dLim[1] = DBL_MAX;
      break;
      case 1:
         ii = 0;                      /* one value for all */
      default:
         if ( pdvMax[ii] != DBL_MAX ) {
            sFP.dLim[1] = max( sFP.dVal, pdvMax[ii] );
         } else {
            if ( *plNoLim ) sFP.dLim[1] = DBL_MAX;
         }
      break;
      }
                                      /* store back parameter values */
      if ( (iRC=fitFctBlkParPutByNum( &sFit.sFFB, plvPar[ip], &sFP, lFlag|FIT_NOMSG ))!= 0 ) goto gError;

   }
/*
   --- list parameters
*/
   if ( *plList ) {
      char cStr[255];
      struct SFITSTAT sFS;
                                      /* if data defined: parameter errors */
      if ( sFit.sFDFit.nn > 0 ) {
                                      /* function derivatives */
         iRC = fitDataFctBlkDeriv(&sFit.sFDFit,&sFit.sFFB,lFlag);
                                      /* curvature matrix */
         iRC = fitDataFctBlkCurv (&sFit.sFDFit,&sFit.sFFB,lFlag | lFlagCurve );
                                      /* covariance matrix */
         iRC = fitFctBlkCovar    (             &sFit.sFFB,lFlag);
                                      /* potential asymmetric errors */
         sFS = sFit.sFFB.sFS;
         sFS.dChisq += 1E0;
         if ( iRC = fitDataFctBlkConfLim( &sFit.sFDFit, &sFit.sFFB, &sFS, lFlag ) )
            goto gError;
      }
                                      /* get header line */
      if ( (iRC=fitFctBlkParStrByNum( &sFit.sFFB, 0, cStr, lFlag|lFlagPeak|FIT_NOMSG )) < 0 ) goto gError;
      _prtcl( cStr, CMDPRTCL_TERM | CMDPRTCL_FILE );
      for ( ip=0; ip<nPar; ip++ ) {
                                      /* get stored parameter values */
         if ( (iRC=fitFctBlkParStrByNum( &sFit.sFFB, plvPar[ip], cStr, lFlag|lFlagPeak|FIT_NOMSG )) < 0 ) goto gError;
         _prtcl( cStr, CMDPRTCL_TERM | CMDPRTCL_FILE ); 
      }
   }
                                      /*  */
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
   if ( plvPeak) free( plvPeak); plvPeak = NULL;
   if ( plvPar ) free( plvPar ); plvPar = NULL;
   if ( pdvGuess ) free( pdvGuess ); pdvGuess = NULL;
   if ( pdvMin ) free( pdvMin ); pdvMin = NULL;
   if ( pdvMax ) free( pdvMax ); pdvMax = NULL;

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return;
}
