#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <limits.h>

#include "u.h"
#include "m.h"
#include "cmd.h"
#include "g.h"
#include "fit.h"

#define FIT_MAXOUT 100

static char rcsid[] = {"$Id: fitcmdres.c,v 1.2 1997/12/15 18:56:09 kraemer Exp $"};

extern struct SFIT sFit;
/*
   --- Command processor for the FRESULT command

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitcmdres.c,v $
   Revision 1.2  1997/12/15 18:56:09  kraemer
   fit statistics in function block

   Revision 1.1.1.1  1997/07/10 18:00:00  kraemer
   revision for cvs

*/

void fitCmdRes( const char *pcFct,
                long *plParam, long *plCorr,
                const char *pcXVal, const char *pcStore,
                double *pdConf,
                long *plGraphic, long *plSigma, long *plStat,
                long *plDebug
               ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__, *pMsg = pDbgMsg;
   unsigned long lFlag = 0, lFlagI = 0, lFlagCurve = 0;

   unsigned long lOut = FIT_OUTLIST;
   unsigned long lDim;
   double *pvdX = NULL;
   int ii, jj, nn;

   int nFct;
   long *pvlFct = NULL;

   struct SFITDATA sFD;
   struct SFITFCTBLK sFFB, *psFFB;
   struct SFITPAR sFP;

   struct List sLTxt;
   char cStr[255+1];
   char *pcStr = NULL;

   int nvItem2D[] = {0,0}, nvItem1D[] = {0};
   double **dvItem2D = NULL;
   double  *dvItem1D = NULL;


   if ( *plDebug ) lFlag |= FIT_DEBUG;
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   if ( lFlag & FIT_DEBUG ) lFlagI |= CMD_DEBUG;
   if ( sFit.lStatus & FIT_POISSON ) lFlagCurve |= FITFCTBLK_POISSON;
   memset( &sFFB, 0, sizeof(sFFB) );
   memset( &sFD, 0, sizeof(sFD) );
   NewList( &sLTxt );
                                      /* check valid inputs */
   while ( *pcXVal == ' ' ) pcXVal++;

   nn = 0;
   if ( *pcXVal  ) nn++;
   if ( *plStat  ) nn++;
   if ( *plCorr  ) nn++;
   if ( *plParam ) nn++;
   if ( nn > 1 ) {
      pMsg = "function(),param,corr,stat";
      iRC = ERRFIT_CMDPAREXCL; goto gError;
   }
   if ( nn <= 0 ) *plParam = 1;
                                      /* output options */
   while ( *pcStore == ' ' ) pcStore++;
   if ( *pcStore ) {
/*
      pMsg = "store()";
      iRC = ERRFIT_CMDPARNOTYET; goto gError;
*/
      lOut = FIT_OUTPAR;
   }
   if ( *plGraphic ) {
                                      /* check graphic text output
                                         interface exists */
      if ( sFit.eFitGfxTxt == NULL ) {
         iRC = ERRFIT_GFXTXTNOTAVAIL; goto gError;
      }
      lOut = FIT_OUTGFX;
   }

   if ( *pdConf >= 1E0 ) {
      iRC = ERRFIT_INVCONF; goto gError;
   }
   if ( *pdConf <= 0E0 ) {
      iRC = ERRFIT_INVCONF; goto gError;
   }
/*
   --- handle function(s)
*/
   while ( *pcFct == ' ' ) pcFct++;
   switch( *pcFct ) {
                                      /* all functions */
   case '\0':
   case '*':
      psFFB = &sFit.sFFB;
   break;
                                      /* selected functions */
   default:
                                      /* count 'em */
      if ( (nFct = c2vl( (char *) pcFct, pvlFct, 1, INT_MAX )) <= 0 ) {
         pMsg = (char *) pcFct; iRC = ERRFIT_CONVNAN; goto gError;
      }
      if ( (pvlFct = calloc( nFct, sizeof(*pvlFct) ) ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
                                      /* store 'em */
      if ( ( c2vl( (char *) pcFct, pvlFct, 1, nFct ) ) != nFct ) {
         pMsg = (char *) pcFct; iRC = ERRFIT_CONVNAN; goto gError;
      }
                                      /* build new function block */
      if ( ( iRC = fitFctBlkByNum( &sFit.sFFB, pvlFct, nFct, &sFFB, lFlag | FIT_NOMSG ) ) != 0 )
         goto gError;
                                      /* remember actual function block */
      psFFB = &sFFB;
   break;
   }

   if ( psFFB->nf <= 0 ) {
      iRC = ERRFIT_NOFUNCTION; goto gError;
   }
/*
   --- Function values
*/
   if ( *pcXVal ) {

      if ( sFit.sFDFit.nn <= 0 ) {
         iRC = ERRFIT_NODATA; goto gError;
      }
                                      /* setup x-values */
      if ( ( nn = c2vd( pcXVal, pvdX, 1, INT_MAX ) ) < 0 ) {
         pMsg = (char *) pcXVal; iRC = ERRFIT_CONVNAN; goto gError;
      }
      if ( ( pvdX = calloc( nn, sizeof(double) ) ) == NULL ) {
         pMsg = (char *) pcXVal; iRC = ERRFIT_SYS; goto gError;
      }
      if ( ( nn = c2vd( pcXVal, pvdX, 1, INT_MAX ) ) < 0 ) {
         pMsg = (char *) pcXVal; iRC = ERRFIT_CONVNAN; goto gError;
      }
      if ( iRC = fitDataAlloc( nn, &sFD, lFlag ) ) goto gError;
      vmov( pvdX, 1, sFD.pdD[FIT_DIMX][FITDATA_DATA], 1, sFD.nn );
                                      /* copy from fitted data struct */
      sFD.lFlag = sFit.sFDFit.lFlag;
      memcpy( sFD.lScale, sFit.sFDFit.lScale, sizeof(sFD.lScale) );
      sFD.dAvgBinsize = sFit.sFDFit.dAvgBinsize;
      sFD.sFCalX = sFit.sFDFit.sFCalX;
      memcpy( sFD.sFDE, sFit.sFDFit.sFDE, sizeof(sFD.sFDE) );
                                      /* covariance from current fit */
                                      /* function derivatives */
      iRC = fitDataFctBlkDeriv( &sFit.sFDFit, &sFit.sFFB, lFlag);
                                      /* curvature matrix */
      iRC = fitDataFctBlkCurv ( &sFit.sFDFit, &sFit.sFFB, lFlag | lFlagCurve );
                                      /* covariance matrix */
      iRC = fitFctBlkCovar    (               &sFit.sFFB, lFlag);
                                      /*  */
      if ( iRC = fitDataScale( &sFD, lDim=FIT_DIMX, lFlag ) ) goto gError;
                                      /* function values and errors */
      if ( iRC = fitDataFctBlkValue( &sFD, psFFB,          lFlag ) ) goto gError;
      if ( iRC = fitDataScaleFct   ( &sFD, FITDATA_SCALEFCT, lFlag ) ) goto gError;
      if ( iRC = fitDataFctBlkConf ( &sFD, psFFB, *pdConf, lFlag ) ) goto gError;
                                      /*  */
      switch( lOut ) {
      case FIT_OUTLIST: case FIT_OUTGFX:
         sprintf( cStr, "%12s %12s %10s %10s",
                        "x", "yfit", "-dyfit", "+dyfit"
                );
         _prtcl( cStr, CMDPRTCL_TERM | CMDPRTCL_FILE );
         if ( lOut & FIT_OUTGFX ) {
            if ( iRC = fitTxtXFerPut( &sLTxt, cStr, lFlag ) ) goto gError;
         }
         nn = min( sFD.nn, FIT_MAXOUT );
         for ( ii=0; ii<sFD.nn; ii++ ) {
            sprintf( cStr, "%12G %12G %10G %10G",
                           sFD.pdD[FIT_DIMX][FITDATA_DATA][ii],
                           sFD.pdFS         [FITDATA_DATA][ii],
                           sFD.pdFS         [FITDATA_ERRM][ii],
                           sFD.pdFS         [FITDATA_ERRP][ii]
                   );
            _prtcl( cStr, CMDPRTCL_TERM | CMDPRTCL_FILE | CMDPRTCL_NOTIME );
            if ( lOut & FIT_OUTGFX ) {
               if ( iRC = fitTxtXFerPut( &sLTxt, cStr, lFlag ) ) goto gError;
            }
         }
      break;
      case FIT_OUTPAR:
         nvItem2D[0] = nvItem1D[0] = sFD.nn; nvItem2D[1] = 4;
         if ( (dvItem2D = (double **) dimalloc( sizeof(double), 2, nvItem2D )) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         if ( (dvItem1D = (double  *) dimalloc( sizeof(double), 1, nvItem1D )) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         for ( ii=0; ii<sFD.nn; ii++ ) {
            dvItem2D[ii][0] = sFD.pdD[FIT_DIMX][FITDATA_DATA][ii];
            dvItem2D[ii][1] = dvItem1D[ii] =
                              sFD.pdFS         [FITDATA_DATA][ii];
            dvItem2D[ii][2] = sFD.pdFS         [FITDATA_ERRM][ii];
            dvItem2D[ii][3] = sFD.pdFS         [FITDATA_ERRP][ii];
         }
      break;
      default:
      break;
      }
   }
/*
   --- Statistics
*/
   if ( *plStat ) {
      char cStr[255+1];

      if ( sFit.sFDFit.nn <= 0 ) {
         iRC = ERRFIT_NODATA; goto gError;
      }
      if ( sFit.sFFB.sFS.nDegFree <= 0 ) {
         iRC = ERRFIT_STATNDF; goto gError;
      }
      switch( lOut ) {
      case FIT_OUTLIST: case FIT_OUTGFX:
         if ( iRC = fitStatChi2Prob( &sFit.sFFB.sFS, lFlag ) ) goto gError;
         sprintf( cStr, "chisq/ndf=%G, degrees of freedom=%d, probability=%G",
                        sFit.sFFB.sFS.dChisq / sFit.sFFB.sFS.nDegFree,
                        sFit.sFFB.sFS.nDegFree,
                        sFit.sFFB.sFS.dProb );
         _prtcl( cStr, CMDPRTCL_TERM | CMDPRTCL_FILE );
         if ( lOut & FIT_OUTGFX ) {
            if ( iRC = fitTxtXFerPut( &sLTxt, cStr, lFlag ) ) goto gError;
         }
      break;
      case FIT_OUTPAR:
         nvItem2D[0] = 1; nvItem2D[1] = nvItem1D[0] = 3;
         if ( (dvItem2D = (double **) dimalloc( sizeof(double), 2, nvItem2D )) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         if ( (dvItem1D = (double  *) dimalloc( sizeof(double), 1, nvItem1D )) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         ii = 0;
         dvItem1D[0] = dvItem2D[ii][0] = sFit.sFFB.sFS.dChisq / sFit.sFFB.sFS.nDegFree;
         dvItem1D[1] = dvItem2D[ii][1] = sFit.sFFB.sFS.nDegFree;
         dvItem1D[2] = dvItem2D[ii][2] = sFit.sFFB.sFS.dProb;
      break;
      default:
      break;
      }
   }
/*
   --- Correlation matrix
*/
   if ( *plCorr ) {

      if ( sFit.sFDFit.nn <= 0 ) {
         iRC = ERRFIT_NODATA; goto gError;
      }

      if ( iRC = fitDataFctBlkDeriv( &sFit.sFDFit, psFFB, lFlag ) )
         goto gError;
      if ( iRC = fitDataFctBlkCurv ( &sFit.sFDFit, psFFB, lFlag | lFlagCurve ) )
         goto gError;
      if ( iRC = fitFctBlkCovar    (               psFFB, lFlag ) )
         goto gError;

      switch( lOut ) {

      case FIT_OUTLIST: case FIT_OUTGFX:
         if ( ( iRC = fitFctBlkCovarStr ( psFFB, pcStr, lFlag ) ) < 0 )
            goto gError;
         if ( ( pcStr = calloc( iRC+1, sizeof(char) ) ) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         iRC = fitFctBlkCovarStr ( psFFB, pcStr, lFlag );
         if ( iRC < 0 ) goto gError;
         _prtcl( pcStr, CMDPRTCL_TERM | CMDPRTCL_FILE );
         if ( lOut & FIT_OUTGFX ) {
            if ( iRC = fitTxtXFerPut( &sLTxt, pcStr, lFlag ) ) goto gError;
         }
         free(pcStr); pcStr = NULL;
      break;

      case FIT_OUTPAR:
         nvItem2D[0] = psFFB->np; nvItem2D[1] = psFFB->np;
         if ( (dvItem2D = (double **) dimalloc( sizeof(double), 2, nvItem2D )) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         nvItem1D[0] = psFFB->np * psFFB->np;
         if ( (dvItem1D = (double  *) dimalloc( sizeof(double), 1, nvItem1D )) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         for ( ii=0; ii<nvItem2D[0]; ii++ ) {
            for ( jj=0; jj<nvItem2D[1]; jj++ ) {
               dvItem1D[ ii*psFFB->np + jj ] = 
               dvItem2D[ii][jj] = psFFB->pdCovar[ ii*psFFB->np + jj ] /
                                  sqrt( fabs(
                                  psFFB->pdCovar[ ii*psFFB->np + ii ] *
                                  psFFB->pdCovar[ jj*psFFB->np + jj ]
                                            ) );
            }
         }
      break;
      default:
      break;
      }
   }
/*
   --- Parameter values
*/
   if ( *plParam ) {
      char cStr[255+1]; int ip; long lp;

      if ( *plSigma ) lFlag |= FITFCTBLKPAR_SIGMA;
      if ( sFit.sFDFit.nn > 0 ) {
         struct SFITSTAT sFS;
                                      /* function derivatives */
         iRC = fitDataFctBlkDeriv( &sFit.sFDFit, psFFB, lFlag );
                                      /* curvature matrix */
         iRC = fitDataFctBlkCurv ( &sFit.sFDFit, psFFB, lFlag | lFlagCurve );
                                      /* covariance matrix */
         iRC = fitFctBlkCovar    (               psFFB, lFlag );
                                      /* potential asymmetric errors */
         sFS = sFit.sFFB.sFS; 
         sFS.dChisq += 1E0;
         iRC = fitDataFctBlkConfLim( &sFit.sFDFit, psFFB, &sFS, lFlag );
      }

      switch( lOut ) {

      case FIT_OUTLIST: case FIT_OUTGFX:
                                      /* get header line */
         if ( (iRC=fitFctBlkParStrByNum( psFFB, 0, cStr, lFlag|FIT_NOMSG )) < 0 ) goto gError;
         _prtcl( cStr, CMDPRTCL_TERM | CMDPRTCL_FILE );
         if ( lOut & FIT_OUTGFX ) {
            if ( iRC = fitTxtXFerPut( &sLTxt, cStr, lFlag ) ) goto gError;
         }
         for ( ip=0; ip<psFFB->np; ip++ ) {
                                      /* get stored parameter values */
            lp = ip+1;
            if ( (iRC=fitFctBlkParStrByNum( psFFB, lp, cStr, lFlag|FIT_NOMSG )) < 0 ) goto gError;
            _prtcl( cStr, CMDPRTCL_TERM | CMDPRTCL_FILE ); 
            if ( lOut & FIT_OUTGFX ) {
               if ( iRC = fitTxtXFerPut( &sLTxt, cStr, lFlag ) ) goto gError;
            }
         }
      break;

      case FIT_OUTPAR:
         nvItem2D[0] = nvItem1D[0] = psFFB->np; nvItem2D[1] = 3;
         if ( (dvItem2D = (double **) dimalloc( sizeof(double), 2, nvItem2D )) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         if ( (dvItem1D = (double  *) dimalloc( sizeof(double), 1, nvItem1D )) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         for ( ii=0; ii<psFFB->np; ii++ ) {
            if ( iRC=fitFctBlkParGetByNum( psFFB, ii+1, &sFP, lFlag ) ) goto gError;
            dvItem2D[ii][0] = dvItem1D[ii] = sFP.dVal;
            dvItem2D[ii][1] = sFP.dErr[1];
            dvItem2D[ii][2] = sFP.dErr[0];
         }
      break;
      default:
      break;
      }
   }
                                      /* */
   if ( *plGraphic ) {
      if ( sFit.eFitGfxTxt( &sLTxt, lFlag ) < 0 ) {
         iRC = ERRFIT_GFXTXTFAILED; goto gError;
      }
   }
                                      /*  */
   if ( *pcStore ) {
      int iDim, nDim, ii, jj, iItem, jItem;
      int  nLB[MBTREEDIM], nHB[MBTREEDIM];
      struct SBTREEARRAY sBTA;
      struct SIPAR *pIPar = NULL; double *pd1D; double **pd2D;

      char *pcType, *pcInit, *pcCom;
      char *pcStr = NULL;
      char cStr[80];

                                      /* decompose glopar name */
      if ( ( iRC=btreearr( pcStore, strlen(pcStore), &sBTA ) ) <= 0 ) {
         iRC = ERRFIT_IPARINVNAME; goto gError;
      }
      if ( sBTA.nName <= 0 ) {
         iRC = ERRFIT_IPARINVNAME; goto gError;
      }
      if ( sBTA.lFlag & BTREE_SUBEXPR ) {
         iRC = ERRFIT_IPARINVNAME; goto gError;
      }
      if ( sBTA.lFlag & BTREE_WILD ) {
         iRC = ERRFIT_IPARINVNAME; goto gError;
      }
                                      /* query target glopar */
      pIPar = _iparptr( pcStore, lFlagI, &iRC );

      switch( iRC ) {

      case 0:                         /* glopar exists */
                                      /*  */
         if ( pIPar -> lFlag & IPAR_READONLY ) {
            iRC = ERRFIT_IPARREADONLY; goto gError;
         }
                                      /* check target data type */
         switch( pIPar->lType ) {
         default:
            iRC = ERRFIT_IPARINVTYP; goto gError;
         break;
         case IPAR_DOUBLE:            /* OK */
         break;
         }
                                      /* check target dimension */
         switch( pIPar->sBTA.nDim ) {
         case 0:
            iRC = ERRFIT_IPARINVDIM; goto gError;
         break;
         case 1:
                                      /* bounds */
                                      /* default */
            nLB[0] = pIPar->sBTA.sDim[0].iLB;
            nHB[0] = pIPar->sBTA.sDim[0].iHB;
                                      /* check user specified bounds */
            switch( sBTA.nDim ) {
            case 0: break;            /* use default */
            case 1:

               if ( sBTA.sDim[0].nLB <= 0 ) {
                  sBTA.sDim[0].iLB = sBTA.sDim[0].iHB;
                  sBTA.sDim[0].iHB = pIPar->sBTA.sDim[0].iHB;
               }
               if ( sBTA.sDim[0].iLB < pIPar->sBTA.sDim[0].iLB ) {
                  iRC = ERRFIT_IPARLOBOUND; goto gError;
               }
               if ( sBTA.sDim[0].iLB > pIPar->sBTA.sDim[0].iHB ) {
                  iRC = ERRFIT_IPARLOBOUND; goto gError;
               }
               if ( sBTA.sDim[0].iHB < pIPar->sBTA.sDim[0].iLB ) {
                  iRC = ERRFIT_IPARHIBOUND; goto gError;
               }
               if ( sBTA.sDim[0].iHB > pIPar->sBTA.sDim[0].iHB ) {
                  iRC = ERRFIT_IPARHIBOUND; goto gError;
               }
               nLB[0] = max( nLB[0], sBTA.sDim[0].iLB );
               nHB[0] = min( nHB[0], sBTA.sDim[0].iHB );
            break;
            default:
               iRC = ERRFIT_IPARINVDIM1; goto gError;
            break;
            }

            nHB[0] = min( nLB[0] + nvItem1D[0] - 1, pIPar->sBTA.sDim[0].iHB );

            if ( (nHB[0] - nLB[0] + 1) <= 0 ) {
               iRC = ERRFIT_IPARRANGEZERO; goto gError;
            }
                                      /* copy data */
            pd1D = (double *) pIPar->pv; iItem = 0;
            for ( ii=nLB[0]; ii<=nHB[0]; ii++ ) {
               pd1D[ii] = dvItem1D[iItem];
               iItem++;
            }

         break; /* target 1D */

         case 2:                      /* target 2D */
                                      /* default */
            for ( iDim=0; iDim<pIPar->sBTA.nDim; iDim++ ) {
               nLB[iDim] = pIPar->sBTA.sDim[iDim].iLB;
               nHB[iDim] = pIPar->sBTA.sDim[iDim].iHB;
            }
                                      /* check user specified bounds */
            switch( sBTA.nDim ) {
            case 0: break;            /* use default */
            case 2:

               for ( iDim=0; iDim<sBTA.nDim; iDim++ ) {
                  if ( sBTA.sDim[iDim].nLB <= 0 ) {
                     sBTA.sDim[iDim].iLB = sBTA.sDim[iDim].iHB;
                     sBTA.sDim[iDim].iHB = pIPar->sBTA.sDim[iDim].iHB;
                  }
                  if ( sBTA.sDim[iDim].iLB < pIPar->sBTA.sDim[iDim].iLB ) {
                     iRC = ERRFIT_IPARLOBOUND; goto gError;
                  }
                  if ( sBTA.sDim[iDim].iLB > pIPar->sBTA.sDim[iDim].iHB ) {
                     iRC = ERRFIT_IPARLOBOUND; goto gError;
                  }
                  if ( sBTA.sDim[iDim].iHB < pIPar->sBTA.sDim[iDim].iLB ) {
                     iRC = ERRFIT_IPARHIBOUND; goto gError;
                  }
                  if ( sBTA.sDim[iDim].iHB > pIPar->sBTA.sDim[iDim].iHB ) {
                     iRC = ERRFIT_IPARHIBOUND; goto gError;
                  }
                  nLB[iDim] = max( nLB[iDim], sBTA.sDim[iDim].iLB );
                  nHB[iDim] = min( nHB[iDim], sBTA.sDim[iDim].iHB );
               }
            break;
            default:
               iRC = ERRFIT_IPARINVDIM2; goto gError;
            break;
            }

            for ( iDim=0; iDim<pIPar->sBTA.nDim; iDim++ ) {
               nHB[iDim] = min( nLB[iDim] + nvItem2D[iDim+1] - 1, pIPar->sBTA.sDim[iDim].iHB );
               if ( (nHB[iDim] - nLB[iDim] + 1) <= 0 ) {
                  iRC = ERRFIT_IPARRANGEZERO; goto gError;
               }
            }
                                      /* copy data */
            pd2D = (double **) pIPar->pv; iItem = 0;
            for ( ii=nLB[0]; ii<=nHB[0]; ii++ ) {
               jItem = 0;
               for ( jj=nLB[1]; jj<=nHB[1]; jj++ ) {
                  pd2D[ii][jj] = dvItem2D[iItem][jItem++];
               }
               iItem++;
            }
                                      /* */
         break; /* target dim 2D */

         default:                     /* dim > 2 */
         break;
         }
      break; /* glopar exists */

      case ERRCMD_IPARNOTFOUND:       /* glopar does not exist */
                                      /* build name */
         sprintf( cStr, "(%d,%d)", nvItem2D[0], nvItem2D[1] );
         if ( ( pcStr = strdup2( pcStore, cStr ) ) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
                                      /* create glopar */
         pIPar = _ipardef( pcStr, pcType="double", pcInit="", pcCom="fres",
                           lFlag | IPAR_DYNAMIC, &iRC );
         free(pcStr); pcStr = NULL;
         if ( ( iRC != 0 ) || ( pIPar == NULL ) ) {
            iRC = ERRFIT_IPARCREATEFAILED; goto gError;
         }
                                      /* store values */
         dimcopy( (void **) dvItem2D, pIPar->pv,
                  sizeof(double), nDim=2, nvItem2D );
      break;
      default:
         iRC = ERRFIT_IPARINVNAME; goto gError;
      break;
      }
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup */
   dimfree( (void **) dvItem1D, 1, nvItem1D );
   dimfree( (void **) dvItem2D, 2, nvItem2D );
   if ( pcStr ) free( pcStr ); pcStr = NULL;
   fitDataFree( &sFD, lFlag );
   fitFctBlkFree( &sFFB, lFlag );
   if ( pvlFct ) free(pvlFct); pvlFct = NULL;
   if ( pvdX ) free(pvdX); pvdX = NULL;
   fitTxtXFerFree( &sLTxt, lFlag );

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return;
}
