
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <limits.h>

#include "u.h"
#include "m.h"
#include "btree.h"
#include "g.h"
#include "fit.h"


static char rcsid[] = {"$Id: fitcmdusr.c,v 1.4 2002/11/18 15:22:01 kraemer Exp $"};

extern struct SFIT sFit;

#if defined(AMIGA)
   extern void *mBase;
//    extern void *mcBase;
   extern void *uBase;
//    extern void *dBase;
#endif

/*
   --- Command processor for the FMY command

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitcmdusr.c,v $
   Revision 1.4  2002/11/18 15:22:01  kraemer
   *** empty log message ***

   Revision 1.3  2002/11/18 15:20:47  kraemer
   mcBase ref removed

   Revision 1.2  1997/12/15 18:56:57  kraemer
   check executable external function

   Revision 1.1.1.1  1997/07/10 18:00:02  kraemer
   revision for cvs

*/

void fitCmdUsr( const char *pcName,
                long *plDebug
               ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__, *pMsg = pDbgMsg;
   unsigned long lFlag = 0;

   int np = 0;
   struct SFITFCT sFF, sFFInit;

   char *pcErr = NULL;


   if ( *plDebug ) lFlag |= FIT_DEBUG;
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sFF, 0, sizeof(sFF) );
   memset( &sFFInit, 0, sizeof(sFFInit) );

                                      /* expand file name */
   while ( *pcName == ' ' ) pcName++;
/*
   if ( ( pcFull = uexpfn(pcName) ) == NULL ) goto gError;
*/
                                      /* check if it's an executable file */
   if ( access(pcName,X_OK) == 0 ) {
                                      /* check if it's a loadable function */
      if ( ( sFFInit.hDyn = dynload( pcName ) ) == NULL ) {
         iRC = ERRFIT_CANTLOAD; goto gError;
      }

      sFFInit.eLoad = ( int (*)() ) dynsymbol( sFFInit.hDyn, "usrfct"/*cFct*/ );
      if ( sFFInit.eLoad == NULL ) {
         iRC = ERRFIT_CANTBIND; goto gError;
      }
      sFFInit.eCall = fitusr;
      sFFInit.lType = FITFCT_USR;
#if defined(AMIGA)
      sFFInit.sUsr.uBase = uBase;
      sFFInit.sUsr.mBase = mBase;
//      sFFInit.sUsr.mcBase = mcBase;
      sFFInit.sUsr.LastBase = NULL;
#endif

   } else {                           /* assume an inline formula */
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> inline formula...\n" );
      if ( ( sFFInit.pvBT = c2btree( pcName, strlen(pcName), &pcErr, &iRC ) ) == NULL ) {
         btreeprc( pDbgMsg, pcErr, iRC, BTREE_ERRMSG );
         iRC = ERRFIT_BTREESYNTAX; goto gError;
      }
      if ( ( iRC = btreeprc( sFFInit.pvBT, NULL, 0, BTREE_RESOLVEOP ) ) != 0) {
         btreeprc( pDbgMsg, NULL, iRC, BTREE_ERRMSG );
         iRC = ERRFIT_BTREERESOLVE; goto gError;
      }
      sFFInit.eCall = fitbtree;
      sFFInit.lType = FITFCT_BTREE;
   }
                                      /* query # of parameters */
   if ( ( np = iRC = fitDataFctNumOfPar( &sFit.sFD, &sFFInit, lFlag ) ) <= 0 ) {
      if ( np == 0 ) iRC = ERRFIT_BADPARCOUNT;
      goto gError;
   }
                                      /* allocate local function  */
   if ( ( iRC = fitFctAlloc( np, &sFF, lFlag) ) != 0 ) goto gError;
                                      /* fill entries */
   fitFctInit( &sFF, lFlag);
   switch( sFF.lType = sFFInit.lType ) {
   case FITFCT_USR:
      if ( ( sFF.pcText = strdup3("User function '",pcName,"'") ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
   break;
   case FITFCT_BTREE:
      if ( ( sFF.pcText = strdup3("Inline function '",pcName,"'") ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }     
   break;
   default:
      iRC = ERRFIT_FCTUNKNOWN; goto gError;
   break;
   }
   if ( ( sFF.pcName = strdup(pcName) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
   sFF.eCall = sFFInit.eCall;
   sFF.eLoad = sFFInit.eLoad;
   sFF.hDyn  = sFFInit.hDyn;
   sFF.pvBT  = sFFInit.pvBT;
   sFF.pvBTI = sFFInit.pvBTI;
/*
   sFF.npSpecial = ...;
*/
   sFF.sUsr = sFFInit.sUsr;
   memset( &sFFInit, 0, sizeof(sFFInit) );
                                      /* initialize parameters */
   fitDataFctInitPar( &sFit.sFD, &sFF, lFlag );
                                      /* add to function block */
   sFF.lFlagNoDup |= FITFCT_UNLOAD;   /* enable unload this instance */
   if ( ( iRC = fitFctBlkFctPut( &sFit.sFFB, &sFF, lFlag | FITFCTBLKFCTPUT_DUPENABLE ) ) != 0 ) goto gError;
   sFF.lFlagNoDup &=~FITFCT_UNLOAD;   /* reset */
                                      /*  */
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup */
   fitFctFree( &sFF, lFlag);
   fitFctFree( &sFFInit, lFlag);

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return;
}
