#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "u.h"
#include "m.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitdata.c,v 1.2 1997/12/15 18:57:58 kraemer Exp $"};

/*
   --- Functions to handle fit data

       fitDataGfxLine() : plot a line
       fitDataStatChi() : chi-square etc
       fitDataNetScaled  () : net values, (scaled y) - (actual yfct)
       fitDataNetTrue    () : net values, (true   y) - (backscaled yfct
       fitDataNetTrueNorm() : net values,((true   y) - (backscaled yfct)) / error
       fitDataErr()     : compute data errors
       fitDataScaleFct(): scale function values for lin/log etc
       fitDataScale()   : scale data for lin/log etc
       fitDataFree()    : frees arrays
       fitDataAlloc()   : allocate all arrays
       fitDataInit()    : initialize
       fitDataInspect() : inspect output
       fitDataNodeListFree() : free a list of data nodes
       fitDataNodeFree()     : free a data nodes
 
   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitdata.c,v $
   Revision 1.2  1997/12/15 18:57:58  kraemer
   gfx flushing added

   Revision 1.1.1.1  1997/07/10 18:00:02  kraemer
   revision for cvs

*/

int fitDataGfxLine(
                    struct SFITDATA *psFD,
                    struct SFITGFX  *psFGfx,
                    unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataGfxLine)", *pMsg = pDbgMsg;
   double *pdXY = NULL;
   double *pd[2];


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   if ( psFGfx == NULL ) goto gError;
   if ( psFGfx->pGH == NULL ) goto gError;

   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;

                                      /* contigous array */
   if ( ( pdXY = calloc(psFD->nn,sizeof(double)*2) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }

   vmov( psFD->pdD [FIT_DIMX][FITDATA_DATA], 1,
         pdXY+0, 2, psFD->nn );
                                      /* plot confidence limit */
   if ( lFlag & FITDATAGFXLINE_CONF ) {
      vmov( psFD->pdCS          [FITDATA_ERRP], 1,
            pdXY+1, 2, psFD->nn );
                                      /* sort */
      pd[0] = pdXY+0;
      pd[1] = pdXY+1;
      vshsrtd( pd, 2, 2, psFD->nn );
                                      /* plot line */
      iRC=_GLine( psFGfx->pGH, pdXY, psFD->nn, psFGfx->lFlag );

      vmov( psFD->pdCS          [FITDATA_ERRM], 1,
            pdXY+1, 2, psFD->nn );
                                      /* sort */
      pd[0] = pdXY+0;
      pd[1] = pdXY+1;
      vshsrtd( pd, 2, 2, psFD->nn );
                                      /* plot line */
      iRC=_GLine( psFGfx->pGH, pdXY, psFD->nn, psFGfx->lFlag );

   } else {                           /* plot function */
      vmov( psFD->pdF           [FITDATA_DATA], 1,
            pdXY+1, 2, psFD->nn );
                                      /* sort */
      pd[0] = pdXY+0;
      pd[1] = pdXY+1;
      vshsrtd( pd, 2, 2, psFD->nn );
                                      /* plot line */
      iRC=_GLine( psFGfx->pGH, pdXY, psFD->nn, psFGfx->lFlag );
   }

   _GFlush( psFGfx->pGH, psFGfx->lFlag );
   iRC = 0;

gError:;

   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
   if ( pdXY ) free( pdXY ); pdXY = NULL;

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
double glngam( double );

int fitDataStatChi(
                   struct SFITDATA *psFD,
                   struct SFITSTAT *psFS,
                   unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataStatChi)", *pMsg = pDbgMsg;
   int ii;
   double dW, dd;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   if ( psFS == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;

   if ( psFS -> lStatus & FITSTAT_POISSON ) {
                                      /* accum. likelihood */
      dd = 0E0;
      psFS->dLogLike = 0E0; iRC = ERRFIT_LIKELIHOODFAILED;
      for ( ii=0; ii<psFD->nn; ii++ ) {
         if ( ( psFD->pd[FIT_DIMY][FITDATA_DATA][ii] + 1E0 ) <= 0E0 ) goto gError;
         if ( psFD->pdFF[FITDATA_DATA][ii] <= 0E0 ) goto gError;
/*
         psFS -> dLogLike -= glngam( psFD->pd[FIT_DIMY][FITDATA_DATA][ii] + 1E0 );
*/
         psFS -> dLogLike += log( psFD->pdFF[FITDATA_DATA][ii] ) *
                                  psFD->pd[FIT_DIMY][FITDATA_DATA][ii];
         psFS -> dLogLike -= psFD->pdFF[FITDATA_DATA][ii];
/*
         dd += log(sqrt(psFD->pd[FIT_DIMY][FITDATA_DATA][ii]*(PI+PI)));
*/
      }
      psFS -> dChisq = -psFS->dLogLike*2E0;
   } else {
      psFS -> dChisq = 0E0;
      for ( ii=0; ii<psFD->nn; ii++ ) {
         dW = fabs( psFD->pd[FIT_DIMY][FITDATA_ERRM][ii] ) +
              fabs( psFD->pd[FIT_DIMY][FITDATA_ERRP][ii] );
         if ( (dW *= 0.5E0) == 0E0 ) dW = 1E0;
         else                        dW = 1E0/(dW*dW);
         psFS -> dChisq += ( psFD->pd[FIT_DIMY][FITDATA_DATA][ii] -
                             psFD->pdFF        [FITDATA_DATA][ii] ) *
                           ( psFD->pd[FIT_DIMY][FITDATA_DATA][ii] -
                             psFD->pdFF        [FITDATA_DATA][ii] ) * dW;
      }
      psFS->dLogLike = -psFS->dChisq * 5E-1;
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/

double glngam( double dX ) {

   static double dCTab[] = {
   76.18009173E0, -86.50532033E0, 24.01409822E0, -1.231739516E0,
   0.120858003E-2, -0.536382E-5,
   };
#define MCTAB ( sizeof(dCTab) / sizeof(double) )

   int ii;
   double dXX, dd, dSum;


   if ( (dXX = dX-1E0) < 0E0 ) dXX = -dXX;

   dSum = 1E0; dd = dXX;
   for ( ii=0; ii<MCTAB; ii++ ) {
      dSum += dCTab[ii] / ++dd;
   }

   dSum *= 2.506628275E0;
   dSum = log(dSum) - ( dXX + 5.5E0 ) + log( dXX + 5.5E0 ) * ( dXX + 5E-1 );

   if ( dX < 1E0 ) dSum = log( PI*dXX ) - log( sin(PI*dXX) ) - dSum;
   return(dSum);
}

/*
   =========================================================================
*/
int fitDataNetScaled(
                struct SFITDATA *psFD,
                unsigned long lFlag
              ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataNetScaled)", *pMsg = pDbgMsg;
   unsigned long lAct;
   int ii;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;

                                      /* used data - actual fct */
   vsub( psFD->pd  [FIT_DIMY][FITDATA_DATA], 1,
         psFD->pdFF          [FITDATA_DATA], 1,
         psFD->pdRR[FIT_DIMY][FITDATA_DATA], 1,
         psFD->nn );

                                      /* error propagation */
   for ( ii=0; ii<psFD->nn; ii++ ) {
      psFD->pdRR[FIT_DIMY][FITDATA_ERRM][ii] =
      psFD->pdRR[FIT_DIMY][FITDATA_ERRP][ii] =
        sqrt(
        ( psFD->pdFF          [FITDATA_ERRM][ii] +
          psFD->pdFF          [FITDATA_ERRP][ii] ) *
        ( psFD->pdFF          [FITDATA_ERRM][ii] +
          psFD->pdFF          [FITDATA_ERRP][ii] ) +
        ( psFD->pd  [FIT_DIMY][FITDATA_ERRM][ii] +
          psFD->pd  [FIT_DIMY][FITDATA_ERRP][ii] ) *
        ( psFD->pd  [FIT_DIMY][FITDATA_ERRM][ii] +
          psFD->pd  [FIT_DIMY][FITDATA_ERRP][ii] )
            ) * 5E-1;
   }

   psFD->pdR [FIT_DIMY][FITDATA_DATA] = psFD->pdRR[FIT_DIMY][FITDATA_DATA];
   psFD->pdR [FIT_DIMY][FITDATA_ERRM] = psFD->pdRR[FIT_DIMY][FITDATA_ERRM];
   psFD->pdR [FIT_DIMY][FITDATA_ERRP] = psFD->pdRR[FIT_DIMY][FITDATA_ERRP];

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataNetTrue(
                struct SFITDATA *psFD,
                unsigned long lFlag
              ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataNetTrue)", *pMsg = pDbgMsg;
   unsigned long lAct;
   int ii;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;

                                      /* true data - scaled fct */
   vsub( psFD->pdD [FIT_DIMY][FITDATA_DATA], 1,
         psFD->pdFS          [FITDATA_DATA], 1,
         psFD->pdRR[FIT_DIMY][FITDATA_DATA], 1,
         psFD->nn );
                                      /* error propagation */
   for ( ii=0; ii<psFD->nn; ii++ ) {
      psFD->pdRR[FIT_DIMY][FITDATA_ERRM][ii] =
      psFD->pdRR[FIT_DIMY][FITDATA_ERRP][ii] =
        sqrt(
        ( psFD->pdFS          [FITDATA_ERRM][ii] +
          psFD->pdFS          [FITDATA_ERRP][ii] ) *
        ( psFD->pdFS          [FITDATA_ERRM][ii] +
          psFD->pdFS          [FITDATA_ERRP][ii] ) +
        ( psFD->pdD [FIT_DIMY][FITDATA_ERRM][ii] +
          psFD->pdD [FIT_DIMY][FITDATA_ERRP][ii] ) *
        ( psFD->pdD [FIT_DIMY][FITDATA_ERRM][ii] +
          psFD->pdD [FIT_DIMY][FITDATA_ERRP][ii] )
            ) * 5E-1;
   }

   psFD->pdR [FIT_DIMY][FITDATA_DATA] = psFD->pdRR[FIT_DIMY][FITDATA_DATA];
   psFD->pdR [FIT_DIMY][FITDATA_ERRM] = psFD->pdRR[FIT_DIMY][FITDATA_ERRM];
   psFD->pdR [FIT_DIMY][FITDATA_ERRP] = psFD->pdRR[FIT_DIMY][FITDATA_ERRP];

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataNetTrueNorm(
                    struct SFITDATA *psFD,
                    unsigned long lFlag
                   ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataNetTrueNorm)", *pMsg = pDbgMsg;
   unsigned long lAct;
   double dd;
   int ii;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;

                                      /* (true data) - (scaled fct) */
   if ( iRC = fitDataNetTrue(psFD,lFlag) ) goto gError;
                                      /* normalize with errors */
   for ( ii=0; ii<psFD->nn; ii++ ) {
      dd = ( 
      psFD->pdRR[FIT_DIMY][FITDATA_ERRM][ii] +
      psFD->pdRR[FIT_DIMY][FITDATA_ERRP][ii]
           ) * 5E-1;
      if ( dd == 0E0 ) dd = 1E0;
      psFD->pdRR[FIT_DIMY][FITDATA_DATA][ii] /= dd;
      psFD->pdRR[FIT_DIMY][FITDATA_ERRM][ii] =
      psFD->pdRR[FIT_DIMY][FITDATA_ERRP][ii] = 1E0;
   }

   psFD->pdR [FIT_DIMY][FITDATA_DATA] = psFD->pdRR[FIT_DIMY][FITDATA_DATA];
   psFD->pdR [FIT_DIMY][FITDATA_ERRM] = psFD->pdRR[FIT_DIMY][FITDATA_ERRM];
   psFD->pdR [FIT_DIMY][FITDATA_ERRP] = psFD->pdRR[FIT_DIMY][FITDATA_ERRP];

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataErr(
                 struct SFITDATA *psFD,
                 long lDim,
                 unsigned long lFlag
              ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataErr)", *pMsg = pDbgMsg;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;

   switch(lDim) {
   case FIT_DIMX:
   break;
   case FIT_DIMY:
   break;
   default:
      iRC = ERRFIT_BADDIM; goto gError;
   break;
   }

   iRC = ERRFIT_BADADDR;
   if ( psFD->pdF      [FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pdO[lDim][FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pdD[lDim][FITDATA_DATA] == NULL ) goto gError;
   if ( psFD->pdD[lDim][FITDATA_ERRM] == NULL ) goto gError;
   if ( psFD->pdD[lDim][FITDATA_ERRP] == NULL ) goto gError;

   switch( psFD->sFDE[lDim].lMode ) {
                                      /* reset errors */
   case FITDATAERR_NONE:
      memset( psFD->pdD[lDim][FITDATA_ERRM], 0, psFD->nn*sizeof(double) );
      memset( psFD->pdD[lDim][FITDATA_ERRP], 0, psFD->nn*sizeof(double) );
   break;
                                      /* */
   case FITDATAERR_OWN:
      vmov ( psFD->pdO[lDim][FITDATA_ERRM], 1,
             psFD->pdD[lDim][FITDATA_ERRM], 1,
             psFD->nn );
      vmov ( psFD->pdO[lDim][FITDATA_ERRP], 1,
             psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->nn );
   break;
                                      /* statistical errors */
   case FITDATAERR_STAT:
      vabs ( psFD->pdD[lDim][FITDATA_DATA], 1,
             psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->nn );
      vsqrt( psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->pdD[lDim][FITDATA_ERRM], 1,
             psFD->nn );
      vsqrt( psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->nn );
   break;
                                      /* statistical errors */
   case FITDATAERR_PERCENT:
      vabs ( psFD->pdD[lDim][FITDATA_DATA], 1,
             psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->nn );
      vsmul( psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->sFDE[lDim].dPercent * 1E-2,
             psFD->pdD[lDim][FITDATA_ERRM], 1,
             psFD->nn );
      vsmul( psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->sFDE[lDim].dPercent * 1E-2,
             psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->nn );
   break;
   case FITDATAERR_THEORY:
      vabs ( psFD->pdF      [FITDATA_DATA], 1,
             psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->nn );
      vsqrt( psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->pdD[lDim][FITDATA_ERRM], 1,
             psFD->nn );
      vsqrt( psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->pdD[lDim][FITDATA_ERRP], 1,
             psFD->nn );
   break;
   default:
   break;
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataScaleFct(
                    struct SFITDATA *psFD,
                    unsigned long lAct,
                    unsigned long lFlag
                   ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataScaleFct)", *pMsg = pDbgMsg;
   int ii;
   long lDim;
   double *pd [FITDATA_MITEM], *pdS[FITDATA_MITEM];


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( pd, 0, sizeof(pd) );
   memset( pdS, 0, sizeof(pdS) );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
                                      /* select action / set pointers */
   switch( lAct ) {
   case FITDATA_SCALEFCT:
      pd [FITDATA_DATA] = psFD->pdFF[FITDATA_DATA];
      pd [FITDATA_ERRM] = psFD->pdFF[FITDATA_ERRM];
      pd [FITDATA_ERRP] = psFD->pdFF[FITDATA_ERRP];
      pdS[FITDATA_DATA] = psFD->pdFS[FITDATA_DATA];
      pdS[FITDATA_ERRM] = psFD->pdFS[FITDATA_ERRM];
      pdS[FITDATA_ERRP] = psFD->pdFS[FITDATA_ERRP];
      psFD->pdF [FITDATA_DATA] = psFD->pdFF[FITDATA_DATA];
      if ( lFlag & FITDATA_SCALEERR ) {
         psFD->pdF [FITDATA_ERRM] = psFD->pdFF[FITDATA_ERRM];
         psFD->pdF [FITDATA_ERRP] = psFD->pdFF[FITDATA_ERRP];
      }
   break;
   case FITDATA_SCALEDIFF:
      pd [FITDATA_DATA] = psFD->pdRR[FIT_DIMY][FITDATA_DATA];
      pd [FITDATA_ERRM] = psFD->pdRR[FIT_DIMY][FITDATA_ERRM];
      pd [FITDATA_ERRP] = psFD->pdRR[FIT_DIMY][FITDATA_ERRP];
      pdS[FITDATA_DATA] = psFD->pdRS[FIT_DIMY][FITDATA_DATA];
      pdS[FITDATA_ERRM] = psFD->pdRS[FIT_DIMY][FITDATA_ERRM];
      pdS[FITDATA_ERRP] = psFD->pdRS[FIT_DIMY][FITDATA_ERRP];

   break;
   default:
      iRC = ERRFIT_SCALEINVACT; goto gError;
   break;
   }
                                      /* move data by default */
   vmov  ( pd [FITDATA_DATA], 1,
           pdS[FITDATA_DATA], 1,
           psFD->nn );
   if ( lFlag & FITDATA_SCALEERR ) {
      vmov  ( pd [FITDATA_ERRM], 1,
              pdS[FITDATA_ERRM], 1,
              psFD->nn );
      vmov  ( pd [FITDATA_ERRP], 1,
              pdS[FITDATA_ERRP], 1,
              psFD->nn );
   }
                                      /* representation: scale data */
   if ( psFD -> lFlag & FITDATA_REPR ) {

      switch( psFD->lScale[ lDim=FIT_DIMY ] ) {
                                      /* log scale */
      case FITSCALE_LOG:
         v10tox( pd [FITDATA_DATA], 1,
                 pdS[FITDATA_DATA], 1,
                 psFD->nn );
                                      /* errors too */
         if ( lFlag & FITDATA_SCALEERR ) {

            vmul ( pd [FITDATA_ERRP], 1,
                   pdS[FITDATA_DATA], 1,
                   pdS[FITDATA_ERRP], 1,
                   psFD->nn );
            vsmul( pdS[FITDATA_ERRP], 1,
                   log(1E1),
                   pdS[FITDATA_ERRP], 1,
                   psFD->nn );
            vmul ( pd [FITDATA_ERRM], 1,
                   pdS[FITDATA_DATA], 1,
                   pdS[FITDATA_ERRM], 1,
                   psFD->nn );
            vsmul( pdS[FITDATA_ERRM], 1,
                   log(1E1),
                   pdS[FITDATA_ERRM], 1,
                   psFD->nn );

         }
      break;
                                      /* sqrt scale */
      case FITSCALE_SQRT:
         vspowi( pd [FITDATA_DATA], 1,
                 2,
                 pdS[FITDATA_DATA], 1,
                 psFD->nn );
                                      /* errors too */
         if ( lFlag & FITDATA_SCALEERR ) {

            vmul ( pd [FITDATA_ERRP], 1,
                   pd [FITDATA_DATA], 1,
                   pdS[FITDATA_ERRP], 1,
                   psFD->nn );
            vsmul( pdS[FITDATA_ERRP], 1,
                   2E0,
                   pdS[FITDATA_ERRP], 1,
                   psFD->nn );
            vmul ( pd [FITDATA_ERRM], 1,
                   pd [FITDATA_DATA], 1,
                   pdS[FITDATA_ERRM], 1,
                   psFD->nn );
            vsmul( pdS[FITDATA_ERRM], 1,
                   2E0,
                   pdS[FITDATA_ERRM], 1,
                   psFD->nn );

         }
      break;
      case FITSCALE_RECIP:
         vspowi( pd [FITDATA_DATA], 1,
                 -1,
                 pdS[FITDATA_DATA], 1,
                 psFD->nn );
                                      /* errors too */
         if ( lFlag & FITDATA_SCALEERR ) {
            vmul  ( pd [FITDATA_ERRM], 1,
                    pdS[FITDATA_DATA], 1,
                    pdS[FITDATA_ERRM], 1,
                    psFD->nn );
            vmul  ( pdS[FITDATA_ERRM], 1,
                    pdS[FITDATA_DATA], 1,
                    pdS[FITDATA_ERRM], 1,
                    psFD->nn );
            vneg  ( pdS[FITDATA_ERRM], 1,
                    pdS[FITDATA_ERRM], 1,
                    psFD->nn );
            vmul  ( pd [FITDATA_ERRP], 1,
                    pdS[FITDATA_DATA], 1,
                    pdS[FITDATA_ERRP], 1,
                    psFD->nn );
            vmul  ( pdS[FITDATA_ERRP], 1,
                    pdS[FITDATA_DATA], 1,
                    pdS[FITDATA_ERRP], 1,
                    psFD->nn );
            vneg  ( pdS[FITDATA_ERRP], 1,
                    pdS[FITDATA_ERRP], 1,
                    psFD->nn );
         }
      break;
      default:
         vmov  ( pd [FITDATA_DATA], 1,
                 pdS[FITDATA_DATA], 1,
                 psFD->nn );
         if ( lFlag & FITDATA_SCALEERR ) {
            vmov  ( pd [FITDATA_ERRM], 1,
                    pdS[FITDATA_ERRM], 1,
                    psFD->nn );
            vmov  ( pd [FITDATA_ERRP], 1,
                    pdS[FITDATA_ERRP], 1,
                    psFD->nn );
         }
      break;
      }
                                      /* set current pointers */
      switch( lAct ) {
      case FITDATA_SCALEFCT:
         psFD->pdF[FITDATA_DATA] = pdS[FITDATA_DATA];
         if ( lFlag & FITDATA_SCALEERR ) {
            psFD->pdF[FITDATA_ERRM] = pdS[FITDATA_ERRM];
            psFD->pdF[FITDATA_ERRP] = pdS[FITDATA_ERRP];
         }
      break;
      default: break;
      }
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataScale(
                 struct SFITDATA *psFD,
                 long lDim,
                 unsigned long lFlag
                ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataScale)", *pMsg = pDbgMsg;
   int ii;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;

   switch(lDim) {
   case FIT_DIMX:
                                      /* apply x-calibr., set pointers */
      vsmsa( psFD->pdD [lDim][FITDATA_DATA], 1,
             psFD->sFCalX.dCoeff[1], psFD->sFCalX.dCoeff[0],
             psFD->pdDS[lDim][FITDATA_DATA], 1,
             psFD->nn );
      psFD->pd[lDim][FITDATA_DATA] = psFD->pdDS[lDim][FITDATA_DATA];
      if ( lFlag & FITDATA_SCALEERR ) {
         vsmsa( psFD->pdD [lDim][FITDATA_ERRM], 1,
                psFD->sFCalX.dCoeff[1], psFD->sFCalX.dCoeff[0],
                psFD->pdDS[lDim][FITDATA_ERRM], 1,
                psFD->nn );
         vsmsa( psFD->pdD [lDim][FITDATA_ERRP], 1,
                psFD->sFCalX.dCoeff[1], psFD->sFCalX.dCoeff[0],
                psFD->pdDS[lDim][FITDATA_ERRP], 1,
                psFD->nn );
         psFD->pd[lDim][FITDATA_ERRM] = psFD->pdDS[lDim][FITDATA_ERRM];
         psFD->pd[lDim][FITDATA_ERRP] = psFD->pdDS[lDim][FITDATA_ERRP];
      }
   break;
   case FIT_DIMY:
                                      /* set current pointers */
      psFD->pd[lDim][FITDATA_DATA] = psFD->pdD [lDim][FITDATA_DATA];
      if ( lFlag & FITDATA_SCALEERR ) {
         psFD->pd[lDim][FITDATA_ERRM] = psFD->pdD [lDim][FITDATA_ERRM];
         psFD->pd[lDim][FITDATA_ERRP] = psFD->pdD [lDim][FITDATA_ERRP];
      }
   break;
   default:
      iRC = ERRFIT_BADDIM; goto gError;
   break;
   }
                                      /* representation: scale data */
   if ( psFD -> lFlag & FITDATA_REPR ) {

      switch( psFD->lScale[ lDim ] ) {
                                      /* log scale */
      case FITSCALE_LOG:
         vlog10( psFD->pd  [lDim][FITDATA_DATA], 1,
                 psFD->pdDS[lDim][FITDATA_DATA], 1,
                 psFD->nn );
                                      /* errors too */
         if ( lFlag & FITDATA_SCALEERR ) {
            vdiv ( psFD->pdD[lDim][FITDATA_ERRP], 1,
                   psFD->pdD[lDim][FITDATA_DATA], 1,
                   psFD->pdDS[lDim][FITDATA_ERRP], 1,
                   psFD->nn );
            vsmul( psFD->pdDS[lDim][FITDATA_ERRP], 1,
                   1E0 / log(1E1),
                   psFD->pdDS[lDim][FITDATA_ERRP], 1,
                   psFD->nn );
            vdiv ( psFD->pdD[lDim][FITDATA_ERRM], 1,
                   psFD->pdD[lDim][FITDATA_DATA], 1,
                   psFD->pdDS[lDim][FITDATA_ERRM], 1,
                   psFD->nn );
            vsmul( psFD->pdDS[lDim][FITDATA_ERRM], 1,
                   1E0 / log(1E1),
                   psFD->pdDS[lDim][FITDATA_ERRM], 1,
                   psFD->nn );
         }
      break;
                                      /* sqrt scale */
      case FITSCALE_SQRT:
         vsqrt ( psFD->pd  [lDim][FITDATA_DATA], 1,
                 psFD->pdDS[lDim][FITDATA_DATA], 1,
                 psFD->nn );
                                      /* errors too */
         if ( lFlag & FITDATA_SCALEERR ) {
            vdiv ( psFD->pdD[lDim][FITDATA_ERRP], 1,
                   psFD->pdDS[lDim][FITDATA_DATA], 1,
                   psFD->pdDS[lDim][FITDATA_ERRP], 1,
                   psFD->nn );
            vsmul( psFD->pdDS[lDim][FITDATA_ERRP], 1,
                   5E-1,
                   psFD->pdDS[lDim][FITDATA_ERRP], 1,
                   psFD->nn );
            vdiv ( psFD->pdD[lDim][FITDATA_ERRM], 1,
                   psFD->pdDS[lDim][FITDATA_DATA], 1,
                   psFD->pdDS[lDim][FITDATA_ERRM], 1,
                   psFD->nn );
            vsmul( psFD->pdDS[lDim][FITDATA_ERRM], 1,
                   5E-1,
                   psFD->pdDS[lDim][FITDATA_ERRM], 1,
                   psFD->nn );
         }
      break;
      case FITSCALE_RECIP:
         vspowi( psFD->pd  [lDim][FITDATA_DATA], 1,
                 -1,
                 psFD->pdDS[lDim][FITDATA_DATA], 1,
                 psFD->nn );
                                      /* errors too */
         if ( lFlag & FITDATA_SCALEERR ) {
            vmul  ( psFD->pdD[lDim][FITDATA_ERRM], 1,
                    psFD->pdDS[lDim][FITDATA_DATA], 1,
                    psFD->pdDS[lDim][FITDATA_ERRM], 1,
                    psFD->nn );
            vmul  ( psFD->pdDS[lDim][FITDATA_ERRM], 1,
                    psFD->pdDS[lDim][FITDATA_DATA], 1,
                    psFD->pdDS[lDim][FITDATA_ERRM], 1,
                    psFD->nn );
            vneg  ( psFD->pdDS[lDim][FITDATA_ERRM], 1,
                    psFD->pdDS[lDim][FITDATA_ERRM], 1,
                    psFD->nn );
            vmul  ( psFD->pdD[lDim][FITDATA_ERRP], 1,
                    psFD->pdDS[lDim][FITDATA_DATA], 1,
                    psFD->pdDS[lDim][FITDATA_ERRP], 1,
                    psFD->nn );
            vmul  ( psFD->pdDS[lDim][FITDATA_ERRP], 1,
                    psFD->pdDS[lDim][FITDATA_DATA], 1,
                    psFD->pdDS[lDim][FITDATA_ERRP], 1,
                    psFD->nn );
            vneg  ( psFD->pdDS[lDim][FITDATA_ERRP], 1,
                    psFD->pdDS[lDim][FITDATA_ERRP], 1,
                    psFD->nn );
         }
      break;
      default:
         vmov  ( psFD->pd  [lDim][FITDATA_DATA], 1,
                 psFD->pdDS[lDim][FITDATA_DATA], 1,
                 psFD->nn );
         if ( lFlag & FITDATA_SCALEERR ) {
            vmov  ( psFD->pdD[lDim][FITDATA_ERRM], 1,
                    psFD->pdDS[lDim][FITDATA_ERRM], 1,
                    psFD->nn );
            vmov  ( psFD->pdD[lDim][FITDATA_ERRP], 1,
                    psFD->pdDS[lDim][FITDATA_ERRP], 1,
                    psFD->nn );
         }
      break;
      }
                                      /* set current pointers */
      psFD->pd[lDim][FITDATA_DATA] = psFD->pdDS[lDim][FITDATA_DATA];
      if ( lFlag & FITDATA_SCALEERR ) {
         psFD->pd[lDim][FITDATA_ERRM] = psFD->pdDS[lDim][FITDATA_ERRM];
         psFD->pd[lDim][FITDATA_ERRP] = psFD->pdDS[lDim][FITDATA_ERRP];
      }
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataFree(
                 struct SFITDATA *psFD,
                 unsigned long lFlag
               ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataFree)", *pMsg = pDbgMsg;
   int iDim, ii;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
                                      /* free data arrays */
   for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
      for ( ii=0; ii<FITDATA_MITEM; ii++ ) {
         psFD -> pd[iDim][ii] = NULL;
         if ( psFD->pdRR[iDim][ii] ) free( psFD->pdRR[iDim][ii] ); psFD->pdRR[iDim][ii] = NULL;
         if ( psFD->pdRS[iDim][ii] ) free( psFD->pdRS[iDim][ii] ); psFD->pdRS[iDim][ii] = NULL;
         if ( psFD->pdD [iDim][ii] ) free( psFD->pdD [iDim][ii] ); psFD->pdD [iDim][ii] = NULL;
         if ( psFD->pdO [iDim][ii] ) free( psFD->pdO [iDim][ii] ); psFD->pdO [iDim][ii] = NULL;
         if ( psFD->pdDS[iDim][ii] ) free( psFD->pdDS[iDim][ii] ); psFD->pdDS[iDim][ii] = NULL;
      }
   }

   for ( ii=0; ii<FITDATA_MITEM; ii++ ) {
      psFD -> pdF[ii] = NULL;
      if ( psFD->pdFF[ii] ) free( psFD->pdFF[ii] ); psFD->pdFF[ii] = NULL;
      if ( psFD->pdFS[ii] ) free( psFD->pdFS[ii] ); psFD->pdFS[ii] = NULL;
      psFD -> pdC[ii] = NULL;
      if ( psFD->pdCC[ii] ) free( psFD->pdCC[ii] ); psFD->pdCC[ii] = NULL;
      if ( psFD->pdCS[ii] ) free( psFD->pdCS[ii] ); psFD->pdCS[ii] = NULL;
   }

   for ( ii=0; ii<FITDATA_MAUX; ii++ ) {
      if ( psFD->pdAux[ii] ) free( psFD->pdAux[ii] ); psFD->pdAux[ii] = NULL;
   }

   psFD->nn = 0;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataAlloc( int nn,
                  struct SFITDATA *psFD,
                  unsigned long lFlag
                ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataAlloc)", *pMsg = pDbgMsg;
   struct SFITDATA sFD;
   int iDim, ii;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sFD, 0, sizeof(sFD) );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( nn <= 0 ) goto gError;

   iRC = ERRFIT_SYS;
                                      /* allocate data arrays / set defaults */
   for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
      for ( ii=0; ii<FITDATA_MITEM; ii++ ) {
         if ( ( sFD.pdRR[iDim][ii] = calloc(nn,sizeof(double)) ) == NULL )
            goto gError;
         if ( ( sFD.pdRS[iDim][ii] = calloc(nn,sizeof(double)) ) == NULL )
            goto gError;
         if ( ( sFD.pdD [iDim][ii] = calloc(nn,sizeof(double)) ) == NULL )
            goto gError;
         if ( ( sFD.pdO [iDim][ii] = calloc(nn,sizeof(double)) ) == NULL )
            goto gError;
         if ( ( sFD.pdDS[iDim][ii] = calloc(nn,sizeof(double)) ) == NULL )
            goto gError;
         sFD.pd[iDim][ii] = sFD.pdD[iDim][ii];
      }
   }
   for ( ii=0; ii<FITDATA_MITEM; ii++ ) {
      if ( ( sFD.pdFF[ii] = calloc(nn,sizeof(double)) ) == NULL )
         goto gError;
      if ( ( sFD.pdFS[ii] = calloc(nn,sizeof(double)) ) == NULL )
         goto gError;
      sFD.pdF[ii] = sFD.pdFF[ii];
      if ( ( sFD.pdCC[ii] = calloc(nn,sizeof(double)) ) == NULL )
         goto gError;
      if ( ( sFD.pdCS[ii] = calloc(nn,sizeof(double)) ) == NULL )
         goto gError;
      sFD.pdC[ii] = sFD.pdCC[ii];
   }
   for ( ii=0; ii<FITDATA_MAUX; ii++ ) {
      if ( ( sFD.pdAux[ii] = calloc(nn,sizeof(double)) ) == NULL )
         goto gError;
   }

   sFD.nn = nn;
/*
   fitDataInit( &sFD, lFlag );
*/
   sFD.dAvgBinsize = 1E0;
   fitCalInit( &sFD.sFCalX, lFlag );

   *psFD = sFD; memset( &sFD, 0, sizeof(sFD) );
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup in case of errors */
   if ( iRC < 0 ) fitDataFree( &sFD, lFlag );

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataInit( struct SFITDATA *psFD,
              unsigned long lFlag
            ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataInit)", *pMsg = pDbgMsg;
   int iDim, ii;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;

   psFD->lFlag = 0;
   memset( psFD->lScale, 0, sizeof(psFD->lScale) );
   psFD->dAvgBinsize = 1E0;
   fitCalInit( &psFD->sFCalX, lFlag );
   psFD->sFDE[0].lMode = 0 /*FITDATAERR_NONE*/;
   psFD->sFDE[0].dPercent = 0E0;
   psFD->sFDE[1].lMode = 0 /*FITDATAERR_NONE*/;
   psFD->sFDE[1].dPercent = 0E0;

   if ( psFD->nn > 0 ) {
      for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
         for ( ii=0; ii<FITDATA_MITEM; ii++ ) {
            if ( psFD->pdRR[iDim][ii] ) memset( psFD->pdRR[iDim][ii] , 0, psFD->nn * sizeof(double) );
            if ( psFD->pdRS[iDim][ii] ) memset( psFD->pdRS[iDim][ii] , 0, psFD->nn * sizeof(double) );
            if ( psFD->pdD [iDim][ii] ) memset( psFD->pdD [iDim][ii] , 0, psFD->nn * sizeof(double) );
            if ( psFD->pdO [iDim][ii] ) memset( psFD->pdO [iDim][ii] , 0, psFD->nn * sizeof(double) );
            if ( psFD->pdDS[iDim][ii] ) memset( psFD->pdDS[iDim][ii] , 0, psFD->nn * sizeof(double) );
         }
      }
      for ( ii=0; ii<FITDATA_MITEM; ii++ ) {
         if ( psFD->pdFF[ii] ) memset( psFD->pdFF[ii] , 0, psFD->nn * sizeof(double) );
         if ( psFD->pdFS[ii] ) memset( psFD->pdFS[ii] , 0, psFD->nn * sizeof(double) );
         if ( psFD->pdCC[ii] ) memset( psFD->pdCC[ii] , 0, psFD->nn * sizeof(double) );
         if ( psFD->pdCS[ii] ) memset( psFD->pdCS[ii] , 0, psFD->nn * sizeof(double) );
      }
      for ( ii=0; ii<FITDATA_MAUX; ii++ ) {
         if ( psFD->pdAux[ii] ) memset( psFD->pdAux[ii] , 0, psFD->nn * sizeof(double) );
      }
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataInspect( struct SFITDATA *psFD,
                    unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataInspect)", *pMsg = pDbgMsg;
   int iDim, iItem, ii;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFD == NULL ) goto gError;

   fprintf( stderr, "<D> SFITDATA=%X\n", psFD );
   fprintf( stderr, "<D> flag=%X\n", psFD->lFlag );
   for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
      fprintf( stderr, "<D> scale[%d]=%d\n", iDim, psFD->lScale[iDim] );
   }
   fprintf( stderr, "<D> nn=%d\n", psFD->nn );
   fprintf( stderr, "<D> binsize=%G\n", psFD->dAvgBinsize );
   fprintf( stderr, "<D> cal=%G %G\n", psFD->sFCalX.dCoeff[0], psFD->sFCalX.dCoeff[1] );
   fprintf( stderr, "<D> data error=%X %G %X %G\n",
                         psFD->sFDE[0].lMode, psFD->sFDE[0].dPercent,
                         psFD->sFDE[1].lMode, psFD->sFDE[1].dPercent
                         );
   for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
   for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
      fprintf( stderr, "<D> pd[%d][%d]=%X pdD[%d][%d]=%X pdDS[%d][%d]=%X\n",
                       iDim, iItem, psFD->pd[iDim][iItem],
                       iDim, iItem, psFD->pdD[iDim][iItem],
                       iDim, iItem, psFD->pdDS[iDim][iItem]
             );
   }
   }
   for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
      fprintf( stderr, "<D> pdF[%d]=%X pdFF[%d]=%X pdFS[%d]=%X\n",
                       iItem, psFD->pdF [iItem],
                       iItem, psFD->pdFF[iItem],
                       iItem, psFD->pdFS[iItem]
             );
   }
   for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
      fprintf( stderr, "<D> pdC[%d]=%X pdCC[%d]=%X pdCS[%d]=%X\n",
                       iItem, psFD->pdC [iItem],
                       iItem, psFD->pdCC[iItem],
                       iItem, psFD->pdCS[iItem]
             );
   }
   for ( iItem=0; iItem<FITDATA_MAUX; iItem++ ) {
      fprintf( stderr, "<D> pdAux[%d]=%X\n",
                       iItem, psFD->pdAux[iItem]
             );
   }

   if ( psFD->pd[FIT_DIMX][FITDATA_DATA] &&
        psFD->pd[FIT_DIMX][FITDATA_ERRM] &&
        psFD->pd[FIT_DIMX][FITDATA_ERRP] &&
        psFD->pd[FIT_DIMY][FITDATA_DATA] &&
        psFD->pd[FIT_DIMY][FITDATA_ERRM] &&
        psFD->pd[FIT_DIMY][FITDATA_ERRP] &&
        psFD->pdF         [FITDATA_DATA] &&
        psFD->pdF         [FITDATA_ERRM] &&
        psFD->pdF         [FITDATA_ERRP] )
      for ( ii=0; ii<psFD->nn; ii++ ) {
         fprintf( stderr, "<D> %G %G %G %G %G %G %G %G %G %G %G\n",
                          psFD->pd[FIT_DIMX][FITDATA_DATA][ii],
                          psFD->pd[FIT_DIMX][FITDATA_ERRM][ii],
                          psFD->pd[FIT_DIMX][FITDATA_ERRP][ii],
                          psFD->pd[FIT_DIMY][FITDATA_DATA][ii],
                          psFD->pd[FIT_DIMY][FITDATA_ERRM][ii],
                          psFD->pd[FIT_DIMY][FITDATA_ERRP][ii],
                          psFD->pdF         [FITDATA_DATA][ii],
                          psFD->pdF         [FITDATA_ERRM][ii],
                          psFD->pdF         [FITDATA_ERRP][ii],
                          psFD->pdRR[FIT_DIMY][FITDATA_DATA][ii],
                          psFD->pdRS[FIT_DIMY][FITDATA_DATA][ii]
                );
      }
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataNodeListFree(
                        struct List *pslh,
                        unsigned long lFlag
                       ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataNodeListFree)", *pMsg = pDbgMsg;
   struct Node *psln;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( pslh == NULL ) goto gError;
                                      /*  */
   while( pslh->lh_Head->ln_Succ ) {
      psln = RemTail( pslh );
      fitDataNodeFree( (struct SFITDATANODE *) psln, lFlag );
      free( (struct SFITDATANODE *) psln);
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataNodeFree(
                    struct SFITDATANODE *psFDN,
                    unsigned long lFlag
                   ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataNodeFree)", *pMsg = pDbgMsg;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFDN == NULL ) goto gError;
                                      /*  */
   if (psFDN->ln.ln_Name) free(psFDN->ln.ln_Name); psFDN->ln.ln_Name = NULL;
   fitDataFree( &psFDN->sFD, lFlag );

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
