#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "m.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitdatawdw.c,v 1.1.1.1 1997/07/10 18:00:03 kraemer Exp $"};

/*
   --- Functions to handle fit data + windows

       fitDataWdw()          : check data against window limits,
                               create new data block
       fitDataWdwList()      : check data against list of windows,
                               create list of data blocks
       fitDataWdwListMerge() : check data against list of windows,
                               create data blocks
       fitDataWdwMom()       : compute moments
 
   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitdatawdw.c,v $
   Revision 1.1.1.1  1997/07/10 18:00:03  kraemer
   revision for cvs

*/

int fitDataWdwListMerge(
                        struct SFITDATA *psFDFr,
                        struct List     *pslhFW,
                        struct SFITDATA *psFDTo,
                        unsigned long lFlag
                       ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataWdwListMerge)", *pMsg = pDbgMsg;
   struct Node *psln;
   struct SFITWDWNODE *psFWN = NULL;
   int ii, nn, iDim, nDim, iItem, iPass, bInside;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFDTo == NULL ) goto gError;
   memset( psFDTo, 0, sizeof(*psFDTo) );
   if ( psFDFr == NULL ) goto gError;
   if ( pslhFW == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFDFr->nn <= 0 ) goto gError;

   for ( iPass=0; iPass<2; iPass++ ) {
      nn = 0;
      for ( ii=0; ii<psFDFr->nn; ii++ ) {
         bInside = 0;
                                      /* loop over input window list */
         for ( psln = pslhFW->lh_Head; psln->ln_Succ; psln=psln->ln_Succ ) {
            psFWN = (struct SFITWDWNODE *) psln;
            nDim = min( psFWN->sFW.nDim, FITDATA_MDIM );

            for ( iDim=0; iDim<nDim; iDim++ ) {
               if ( psFDFr->pdD[iDim][FITDATA_DATA][ii] < psFWN->sFW.dLim[iDim][0] ) goto gSkip;
               if ( psFDFr->pdD[iDim][FITDATA_DATA][ii] > psFWN->sFW.dLim[iDim][1] ) goto gSkip;
            }
            bInside = 1;
gSkip:;
         }
         if ( bInside ) {
            if ( iPass > 0 ) {
               for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
               for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
                  psFDTo -> pdD [iDim][iItem][nn] = psFDFr -> pdD [iDim][iItem][ii];
                  psFDTo -> pdO [iDim][iItem][nn] = psFDFr -> pdO [iDim][iItem][ii];
                  psFDTo -> pdDS[iDim][iItem][nn] = psFDFr -> pdDS[iDim][iItem][ii];
                  psFDTo -> pdRR[iDim][iItem][nn] = psFDFr -> pdRR[iDim][iItem][ii];
                  psFDTo -> pdRS[iDim][iItem][nn] = psFDFr -> pdRS[iDim][iItem][ii];
               }
               }
               for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
                  psFDTo -> pdFF[iItem][nn] = psFDFr -> pdFF[iItem][ii];
                  psFDTo -> pdFS[iItem][nn] = psFDFr -> pdFS[iItem][ii];
                  psFDTo -> pdCC[iItem][nn] = psFDFr -> pdCC[iItem][ii];
                  psFDTo -> pdCS[iItem][nn] = psFDFr -> pdCS[iItem][ii];
               }
            }
            nn++;
         }
      } /* data loop */

      if ( nn <= 0 ) {
         iRC = ERRFIT_WDWEMPTY; goto gError;
      }
      if ( iPass == 0 ) {
         if ( ( iRC = fitDataAlloc( nn, psFDTo, lFlag ) ) != 0 ) goto gError;
      }
   } /* pass */

                                      /* store rest of entries */
   psFDTo -> lFlag = psFDFr -> lFlag;
   for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
      psFDTo -> lScale[iDim] = psFDFr -> lScale[iDim];
      psFDTo -> sFDE  [iDim] = psFDFr -> sFDE  [iDim];
   }
   psFDTo -> dAvgBinsize = psFDFr -> dAvgBinsize;
   psFDTo -> sFCalX = psFDFr -> sFCalX;

   for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
      for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
         if ( psFDFr -> pd[iDim][iItem] == psFDFr -> pdD [iDim][iItem] )
              psFDTo -> pd[iDim][iItem] =  psFDTo -> pdD [iDim][iItem];
         if ( psFDFr -> pd[iDim][iItem] == psFDFr -> pdDS[iDim][iItem] )
              psFDTo -> pd[iDim][iItem] =  psFDTo -> pdDS[iDim][iItem];
         if ( psFDFr -> pdR[iDim][iItem] == psFDFr -> pdRR[iDim][iItem] )
              psFDTo -> pdR[iDim][iItem] =  psFDTo -> pdRR[iDim][iItem];
         if ( psFDFr -> pdR[iDim][iItem] == psFDFr -> pdRS[iDim][iItem] )
              psFDTo -> pdR[iDim][iItem] =  psFDTo -> pdRS[iDim][iItem];
      }
   }
   for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
      if ( psFDFr -> pdF [iItem] == psFDFr -> pdFF[iItem] )
           psFDTo -> pdF [iItem] =  psFDTo -> pdFF[iItem];
      if ( psFDFr -> pdF [iItem] == psFDFr -> pdFS[iItem] )
           psFDTo -> pdF [iItem] =  psFDTo -> pdFS[iItem];
      if ( psFDFr -> pdC [iItem] == psFDFr -> pdCC[iItem] )
           psFDTo -> pdC [iItem] =  psFDTo -> pdCC[iItem];
      if ( psFDFr -> pdC [iItem] == psFDFr -> pdCS[iItem] )
           psFDTo -> pdC [iItem] =  psFDTo -> pdCS[iItem];
   }


   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
   if ( (iRC < 0) && psFDTo ) {
      fitDataFree( psFDTo, lFlag );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataWdwMom (
                    struct SFITDATA *psFD,
                    struct SFITWDW  *psFW,
                    struct SFITMOM  *psFMom,
                    unsigned long lAct,
                    unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataWdwMom)", *pMsg = pDbgMsg;

   double *pd[FITDATA_MDIM][FITDATA_MITEM];
   int ii;
   double dd = 0E0;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFMom == NULL ) goto gError;
   memset( psFMom, 0, sizeof(*psFMom) );
   if ( psFW == NULL ) goto gError;
   if ( psFD == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;

   switch( lAct ) {
   case FITMOM_DATA:
      pd[FIT_DIMX][FITDATA_DATA] = psFD->pdD [FIT_DIMX][FITDATA_DATA];
      pd[FIT_DIMY][FITDATA_DATA] = psFD->pdD [FIT_DIMY][FITDATA_DATA];
      pd[FIT_DIMY][FITDATA_ERRM] = psFD->pdD [FIT_DIMY][FITDATA_ERRM];
      pd[FIT_DIMY][FITDATA_ERRP] = psFD->pdD [FIT_DIMY][FITDATA_ERRP];
   break;
   case FITMOM_FCT:
      pd[FIT_DIMX][FITDATA_DATA] = psFD->pdD [FIT_DIMX][FITDATA_DATA];
      pd[FIT_DIMY][FITDATA_DATA] = psFD->pdFS          [FITDATA_DATA];
      pd[FIT_DIMY][FITDATA_ERRM] = psFD->pdFS          [FITDATA_ERRM];
      pd[FIT_DIMY][FITDATA_ERRP] = psFD->pdFS          [FITDATA_ERRP];
   break;
   case FITMOM_DIFF:
      pd[FIT_DIMX][FITDATA_DATA] = psFD->pdD [FIT_DIMX][FITDATA_DATA];
      pd[FIT_DIMY][FITDATA_DATA] = psFD->pdRR[FIT_DIMY][FITDATA_DATA];
      pd[FIT_DIMY][FITDATA_ERRM] = psFD->pdRR[FIT_DIMY][FITDATA_ERRM];
      pd[FIT_DIMY][FITDATA_ERRP] = psFD->pdRR[FIT_DIMY][FITDATA_ERRP];
   break;
   default:
      iRC = ERRFIT_MOMINVACT; goto gError;
   break;
   }
                                      /* sum and mean value */
   for ( ii=0; ii<psFD->nn; ii++ ) {

      if ( ( psFD->pdD[FIT_DIMX][FITDATA_DATA][ii] >= psFW->dLim[0][0] ) &&
           ( psFD->pdD[FIT_DIMX][FITDATA_DATA][ii] <= psFW->dLim[0][1] ) &&
           ( psFD->pdD[FIT_DIMY][FITDATA_DATA][ii] >= psFW->dLim[1][0] ) &&
           ( psFD->pdD[FIT_DIMY][FITDATA_DATA][ii] <= psFW->dLim[1][1] ) ) {
         psFMom->dMom[FITMOM_SUM ][FITMOM_VAL] += pd[FIT_DIMY][FITDATA_DATA][ii];
         psFMom->dMom[FITMOM_MEAN][FITMOM_VAL] += pd[FIT_DIMY][FITDATA_DATA][ii] *
                                                  pd[FIT_DIMX][FITDATA_DATA][ii];
                                      /* sum errors (symmetrized) */
         dd += ( pd[FIT_DIMY][FITDATA_ERRM][ii] + pd[FIT_DIMY][FITDATA_ERRP][ii] ) * 5E-1 *
               ( pd[FIT_DIMY][FITDATA_ERRM][ii] + pd[FIT_DIMY][FITDATA_ERRP][ii] ) * 5E-1 ;
         psFMom->nn++;
      }
   }
   psFMom->dMom[FITMOM_SUM ][FITMOM_ERRP] =
   psFMom->dMom[FITMOM_SUM ][FITMOM_ERRM] = sqrt(dd);
                                      /* variance and skewness */
   if ( psFMom->dMom[FITMOM_SUM][FITMOM_VAL] != 0E0 ) {
      psFMom->dMom[FITMOM_MEAN][FITMOM_VAL] /= psFMom->dMom[FITMOM_SUM][FITMOM_VAL];
      for ( ii=0; ii<psFD->nn; ii++ ) {
         if ( ( psFD->pdD[FIT_DIMX][FITDATA_DATA][ii] >= psFW->dLim[0][0] ) &&
              ( psFD->pdD[FIT_DIMX][FITDATA_DATA][ii] <= psFW->dLim[0][1] ) &&
              ( psFD->pdD[FIT_DIMY][FITDATA_DATA][ii] >= psFW->dLim[1][0] ) &&
              ( psFD->pdD[FIT_DIMY][FITDATA_DATA][ii] <= psFW->dLim[1][1] ) ) {
            psFMom->dMom[FITMOM_VAR ][FITMOM_VAL] += pd[FIT_DIMY][FITDATA_DATA][ii] *
                                          ( pd[FIT_DIMX][FITDATA_DATA][ii] -
                                            psFMom->dMom[FITMOM_MEAN][FITMOM_VAL] ) *
                                          ( pd[FIT_DIMX][FITDATA_DATA][ii] -
                                            psFMom->dMom[FITMOM_MEAN][FITMOM_VAL] );
            psFMom->dMom[FITMOM_SKEW][FITMOM_VAL] += pd[FIT_DIMY][FITDATA_DATA][ii] *
                                          ( pd[FIT_DIMX][FITDATA_DATA][ii] -
                                            psFMom->dMom[FITMOM_MEAN][FITMOM_VAL] ) *
                                          ( pd[FIT_DIMX][FITDATA_DATA][ii] -
                                            psFMom->dMom[FITMOM_MEAN][FITMOM_VAL] ) *
                                          ( pd[FIT_DIMX][FITDATA_DATA][ii] -
                                            psFMom->dMom[FITMOM_MEAN][FITMOM_VAL] );
         }
      }
      psFMom->dMom[FITMOM_VAR ][FITMOM_VAL] /= psFMom->dMom[FITMOM_SUM][FITMOM_VAL];
      psFMom->dMom[FITMOM_SKEW][FITMOM_VAL] /= psFMom->dMom[FITMOM_SUM][FITMOM_VAL];
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataWdwList(
                    struct SFITDATA *psFD,
                    struct List *pslhFW,
                    struct List *pslhFD,
                    unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataWdwList)", *pMsg = pDbgMsg;
   struct Node *psln;
   struct SFITDATANODE *psFDN = NULL;
   struct SFITWDWNODE *psFWN = NULL;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( pslhFD == NULL ) goto gError;
   NewList( pslhFD );
   if ( psFD == NULL ) goto gError;
   if ( pslhFW == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFD->nn <= 0 ) goto gError;

                                      /* loop over input window list */
   for ( psln = pslhFW->lh_Head; psln->ln_Succ; psln=psln->ln_Succ ) {
      psFWN = (struct SFITWDWNODE *) psln;
                                      /* new data node */
      if ( ( psFDN = calloc( 1, sizeof(*psFDN) ) ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
                                      /* check window match */
      iRC = fitDataWdw( psFD, &psFWN->sFW, &psFDN->sFD, lFlag );
      switch( iRC ) {
      case ERRFIT_WDWEMPTY: break;
      case 0:
         AddTail( pslhFD, &psFDN->ln ); psFDN = NULL;
      break;
      default:
         goto gError;
      break;
      }
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
   if ( (iRC < 0) && pslhFD ) {
      fitDataNodeListFree( pslhFD, lFlag );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataWdw(
                struct SFITDATA *psFDFr,
                struct SFITWDW  *psFW,
                struct SFITDATA *psFDTo,
                unsigned long lFlag
              ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataWdw)", *pMsg = pDbgMsg;
   int ii, iDim, nDim, iItem, nn, bInside;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFDTo == NULL ) goto gError;
   memset( psFDTo, 0, sizeof(*psFDTo) );
   if ( psFDFr == NULL ) goto gError;
   if ( psFW  == NULL ) goto gError;
   iRC = ERRFIT_BADDATACOUNT;
   if ( psFDFr->nn <= 0 ) goto gError;
   nDim = min( psFW->nDim, FITDATA_MDIM );
                                      /* check window match (count) */
   nn = 0;
   for ( ii=0; ii<psFDFr->nn; ii++ ) {
      bInside = 0;
      for ( iDim=0; iDim<nDim; iDim++ ) {
         if ( psFDFr->pdD[iDim][FITDATA_DATA][ii] < psFW->dLim[iDim][0] ) break;
         if ( psFDFr->pdD[iDim][FITDATA_DATA][ii] > psFW->dLim[iDim][1] ) break;
         bInside++;
      }
      if ( bInside == nDim ) nn++;
   }

   if ( nn <= 0 ) {
      iRC = ERRFIT_WDWEMPTY; goto gError;
   }
   if ( ( iRC = fitDataAlloc( nn, psFDTo, lFlag ) ) != 0 ) goto gError;

                                      /* check window match (store) */
   nn = 0;
   for ( ii=0; ii<psFDFr->nn; ii++ ) {
      bInside = 0;
      for ( iDim=0; iDim<nDim; iDim++ ) {
         if ( psFDFr->pdD[iDim][FITDATA_DATA][ii] < psFW->dLim[iDim][0] ) break;
         if ( psFDFr->pdD[iDim][FITDATA_DATA][ii] > psFW->dLim[iDim][1] ) break;
         bInside++;
      }
      if ( bInside == nDim ) {
         for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
         for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
            psFDTo -> pdD [iDim][iItem][nn] = psFDFr -> pdD [iDim][iItem][ii];
            psFDTo -> pdO [iDim][iItem][nn] = psFDFr -> pdO [iDim][iItem][ii];
            psFDTo -> pdDS[iDim][iItem][nn] = psFDFr -> pdDS[iDim][iItem][ii];
            psFDTo -> pdRR[iDim][iItem][nn] = psFDFr -> pdRR[iDim][iItem][ii];
            psFDTo -> pdRS[iDim][iItem][nn] = psFDFr -> pdRS[iDim][iItem][ii];
         }
         }
         for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
            psFDTo -> pdFF[iItem][nn] = psFDFr -> pdFF[iItem][ii];
            psFDTo -> pdFS[iItem][nn] = psFDFr -> pdFS[iItem][ii];
            psFDTo -> pdCC[iItem][nn] = psFDFr -> pdCC[iItem][ii];
            psFDTo -> pdCS[iItem][nn] = psFDFr -> pdCS[iItem][ii];
         }
         nn++;
      }
   }
                                      /* store rest of entries */
   psFDTo -> lFlag = psFDFr -> lFlag;
   for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
      psFDTo -> lScale[iDim] = psFDFr -> lScale[iDim];
      psFDTo -> sFDE  [iDim] = psFDFr -> sFDE  [iDim];
   }
   psFDTo -> dAvgBinsize = psFDFr -> dAvgBinsize;
   psFDTo -> sFCalX = psFDFr -> sFCalX;

   for ( iDim=0; iDim<FITDATA_MDIM; iDim++ ) {
      for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
         if ( psFDFr -> pd[iDim][iItem] == psFDFr -> pdD [iDim][iItem] )
              psFDTo -> pd[iDim][iItem] =  psFDTo -> pdD [iDim][iItem];
         if ( psFDFr -> pd[iDim][iItem] == psFDFr -> pdDS[iDim][iItem] )
              psFDTo -> pd[iDim][iItem] =  psFDTo -> pdDS[iDim][iItem];
         if ( psFDFr -> pdR[iDim][iItem] == psFDFr -> pdRR[iDim][iItem] )
              psFDTo -> pdR[iDim][iItem] =  psFDTo -> pdRR[iDim][iItem];
         if ( psFDFr -> pdR[iDim][iItem] == psFDFr -> pdRS[iDim][iItem] )
              psFDTo -> pdR[iDim][iItem] =  psFDTo -> pdRS[iDim][iItem];
      }
   }
   for ( iItem=0; iItem<FITDATA_MITEM; iItem++ ) {
      if ( psFDFr -> pdF [iItem] == psFDFr -> pdFF[iItem] )
           psFDTo -> pdF [iItem] =  psFDTo -> pdFF[iItem];
      if ( psFDFr -> pdF [iItem] == psFDFr -> pdFS[iItem] )
           psFDTo -> pdF [iItem] =  psFDTo -> pdFS[iItem];
      if ( psFDFr -> pdC [iItem] == psFDFr -> pdCC[iItem] )
           psFDTo -> pdC [iItem] =  psFDTo -> pdCC[iItem];
      if ( psFDFr -> pdC [iItem] == psFDFr -> pdCS[iItem] )
           psFDTo -> pdC [iItem] =  psFDTo -> pdCS[iItem];
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
   if ( (iRC < 0) && psFDTo ) {
      fitDataFree( psFDTo, lFlag );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
