#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "u.h"
#include "m.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitdataxfer.c,v 1.1.1.1 1997/07/10 18:00:04 kraemer Exp $"};

/*
   --- Functions to handle fit data transfer

       fitDataXFerFree()    : frees arrays
       fitDataXFerAlloc()   : allocate all arrays
       fitDataXFerInit()    : initialize
       fitDataXFerInspect() : inspect output
 
   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitdataxfer.c,v $
   Revision 1.1.1.1  1997/07/10 18:00:04  kraemer
   revision for cvs

*/

int fitDataXFerFree(
                    struct SFITDATAXFER *psFDXF,
                    unsigned long lFlag
                   ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataXFerFree)", *pMsg = pDbgMsg;
   int iDim, ii;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFDXF == NULL ) goto gError;
                                      /* free data arrays */
   for ( iDim=0; iDim<FITXFER_MDIM+1; iDim++ ) {
      for ( ii=0; ii<FITXFER_MITEM; ii++ ) {
         if ( psFDXF->pd[iDim][ii] ) free( psFDXF->pd[iDim][ii] ); psFDXF->pd[iDim][ii] = NULL;
      }
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataXFerAlloc( int nn[], int nDim,
                      struct SFITDATAXFER *psFDXF,
                      unsigned long lFlag
                    ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataXFerAlloc)", *pMsg = pDbgMsg;
   struct SFITDATAXFER sFDXF;
   int ii, iDim;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sFDXF, 0, sizeof(sFDXF) );
                                      /* check valid inputs */
   if ( psFDXF == NULL ) goto gError;
   if ( nn == NULL ) goto gError;

   iRC = ERRFIT_BADDIM;
   if ( nDim <= 0 ) goto gError;
   if ( nDim > FITXFER_MDIM ) goto gError;
                                      /* array size each dimension */
   iRC = ERRFIT_BADDATACOUNT;
   sFDXF.pl[nDim] = 1;
   for ( iDim=0; iDim<nDim; iDim++ ) {
      if ( nn[iDim] <= 0 ) goto gError;
      sFDXF.pl[iDim] = nn[iDim];
      sFDXF.pl[nDim] *= nn[iDim];
   }

   iRC = ERRFIT_SYS;
                                      /* allocate data arrays / set defaults */
   for ( iDim=0; iDim<nDim+1; iDim++ ) {
      for ( ii=0; ii<FITXFER_MITEM; ii++ ) {
         if ( ( sFDXF.pd[iDim][ii] = calloc( sFDXF.pl[iDim], sizeof(double)) ) == NULL )
            goto gError;
      }
   }

   sFDXF.nDim = nDim;
/*
   fitDataXFerInit( &sFDXF, lFlag );
*/

   *psFDXF = sFDXF; memset( &sFDXF, 0, sizeof(sFDXF) );
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup in case of errors */
   if ( iRC < 0 ) fitDataXFerFree( &sFDXF, lFlag );

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataXFerInit( struct SFITDATAXFER *psFDXF,
                     unsigned long lFlag
                   ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataXFerInit)", *pMsg = pDbgMsg;
   int iDim, ii;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFDXF == NULL ) goto gError;

                                      /* initialize data arrays / set defaults */
   for ( iDim=0; iDim<FITXFER_MDIM+1; iDim++ ) {
      if ( psFDXF->pl[iDim] > 0 ) {
         for ( ii=0; ii<FITXFER_MITEM; ii++ ) {
            memset( psFDXF->pd[iDim][ii], 0, psFDXF->pl[iDim] * sizeof(double) );
         }
      }
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitDataXFerInspect( struct SFITDATAXFER *psFDXF,
                        unsigned long lFlag
                      ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitDataXFerInspect)", *pMsg = pDbgMsg;

   int iDim, nDim, iItem, ii;
   unsigned long ll, lv[ FITXFER_MDIM ];


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFDXF == NULL ) goto gError;

   fprintf( stderr, "<D> SFITDATAXFER=%X\n", psFDXF );
   fprintf( stderr, "<D> dim=%d\n", psFDXF->nDim );

   memset( lv, 0, sizeof(lv) );
   nDim = psFDXF->nDim;
   for ( ll=0; ll<psFDXF->pl[ nDim ]; ll++ ) {

      fprintf( stderr, "<D> ll=%d", ll );

      for ( iDim=0; iDim<nDim; iDim++ ) {
         fprintf( stderr, " dim=%d %G %G %G", iDim,
                          psFDXF->pd[ iDim ][ FITXFER_DATA ][ lv[iDim] ],
                          psFDXF->pd[ iDim ][ FITXFER_ERRM ][ lv[iDim] ],
                          psFDXF->pd[ iDim ][ FITXFER_ERRP ][ lv[iDim] ]
                );
      }

      fprintf( stderr, " dim=%d %G %G %G\n", nDim,
                       psFDXF->pd[ nDim ][ FITXFER_DATA ][ll],
                       psFDXF->pd[ nDim ][ FITXFER_ERRM ][ll],
                       psFDXF->pd[ nDim ][ FITXFER_ERRP ][ll]
             );

      ++lv[ nDim-1 ];
      for ( iDim=nDim-1; iDim>=0; iDim-- ) {
         if ( lv[iDim] >= psFDXF->pl[ iDim ] ) {
            lv[iDim] = 0; if ( iDim > 0 ) lv[iDim-1] ++;
         }
      }
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
