#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "m.h"
#include "btree.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitfct.c,v 1.1.1.1 1997/07/10 18:00:04 kraemer Exp $"};

/*
   --- Functions to handle function related structures

       fitFctAlloc()   : allocate
       fitFctFree()    : free
       fitFctInit()    : initialize
       fitFctInspect() : inspect output
       fitFctDup()     : duplicate

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitfct.c,v $
   Revision 1.1.1.1  1997/07/10 18:00:04  kraemer
   revision for cvs

*/

int fitFctDup( struct SFITFCT *psFFFr,
               struct SFITFCT *psFFTo,
               unsigned long lFlag
             ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctDup)", *pMsg = pDbgMsg;
   int np;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFFr == NULL ) goto gError;
   if ( psFFTo == NULL ) goto gError;
   iRC = ERRFIT_BADPARCOUNT;
   if ( ( np = psFFFr -> np ) <= 0 ) goto gError;
                                      /* initialize target */
   memset( psFFTo, 0, sizeof(*psFFTo) );
/*
   fitFctInit( psFFTo, lFlag );
*/
                                      /* space for parameter related stuff*/
   if ( ( iRC = fitFctAlloc( np, psFFTo, lFlag ) ) != 0 )
      goto gError;
                                      /* duplicate substructure, if present */
   if ( psFFFr -> sFPD.pd )
      if ( ( iRC = fitParDerivDup( &psFFFr->sFPD, &psFFTo->sFPD, lFlag ) ) != 0 )
         goto gError;

   memcpy( psFFTo -> pdVal,     psFFFr -> pdVal,     np*sizeof(double) );
   memcpy( psFFTo -> pdRefVal,  psFFFr -> pdRefVal,  np*sizeof(double) );
   memcpy( psFFTo -> pdLim,     psFFFr -> pdLim,     np*sizeof(double)*2 );
   memcpy( psFFTo -> pdConfLim, psFFFr -> pdConfLim, np*sizeof(double)*2 );
   memcpy( psFFTo -> plFlag,    psFFFr -> plFlag,    np*sizeof(unsigned long) );
/*
   memcpy( psFFTo -> plRefNum,  psFFFr -> plRefNum,  np*sizeof(long) );
   memcpy( psFFTo -> plRefType, psFFFr -> plRefType, np*sizeof(long) );
*/
   if ( ( psFFTo->pcText = strdup( psFFFr->pcText ) ) == NULL ) goto gError;
   if ( ( psFFTo->pcName = strdup( psFFFr->pcName ) ) == NULL ) goto gError;

   psFFTo -> lStatus = psFFFr -> lStatus;
   psFFTo -> lType   = psFFFr -> lType;
   if ( lFlag & FITFCTDUP_DUPENABLE ) psFFTo -> lFlagNoDup = psFFFr -> lFlagNoDup;
                                      /* !!! */
   psFFTo -> eLoad = psFFFr -> eLoad;
   psFFTo -> hDyn  = psFFFr -> hDyn;
   psFFTo -> eCall = psFFFr -> eCall;
                                      /* binary tree: duplicate */
   if ( psFFTo -> lType == FITFCT_BTREE ) {
      struct SBTREEVIDF *psBTIFr, *psBTITo;
      int nn = 0;
      char *pcErr;
                                      /* duplicate binary tree */
/*
      psFFTo -> pvBT  = psFFFr -> pvBT;
*/
/*
      iRC = btreeprc( psFFFr->pvBT, psFFTo->pvBT, BTREE_DUP );
   or
*/
      if ( ( psFFTo->pvBT = c2btree( psFFTo->pcName, strlen(psFFTo->pcName),
                                     &pcErr, &iRC ) ) == NULL ) {
         btreeprc( pDbgMsg, pcErr, iRC, BTREE_ERRMSG );
         iRC = ERRFIT_BTREESYNTAX; goto gError;
      }
      if ( ( iRC = btreeprc( psFFTo->pvBT, NULL, 0, BTREE_RESOLVEOP ) ) !=0) {
         btreeprc( pDbgMsg, NULL, iRC, BTREE_ERRMSG );
         iRC = ERRFIT_BTREERESOLVE; goto gError;
      }
                                      /* duplicate identifier table */
/*
      psFFTo -> pvBTI = psFFFr -> pvBTI;
*/
      if ( psBTIFr = psFFFr -> pvBTI ) {
         nn = 0; while( psBTIFr->cIdf ) { psBTIFr++; nn++; }
         if ( ( psFFTo -> pvBTI = calloc( nn+1, sizeof(struct SBTREEVIDF) ) ) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
         psBTIFr = psFFFr -> pvBTI;
         psBTITo = psFFTo -> pvBTI;
         while( psBTIFr->cIdf ) {
            *psBTITo = *psBTIFr;
            if ( (psBTITo -> cIdf = strdup( psBTIFr -> cIdf )) == NULL ) {
               iRC = ERRFIT_SYS; goto gError;
            }
            psBTIFr++; psBTITo++;
         }
/*
   or
         iRC = btreeprc( psFFFr->pvBTI, psFFTo->pvBTI, BTREE_DUPVIDF );
*/
      }
   }

   psFFTo -> sUsr = psFFFr -> sUsr;
                                      /* remember # of parameters */
   psFFTo -> np = np;

   psFFTo -> npSpecial = psFFFr -> npSpecial;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup in case of error */
   if ( ( iRC < 0 ) && psFFTo ) {
      fitFctFree( psFFTo, lFlag );
   }

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/

int fitFctAlloc( int np,
                 struct SFITFCT *psFF,
                 unsigned long lFlag
               ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctAlloc)", *pMsg = pDbgMsg;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFF == NULL ) goto gError;
   iRC = ERRFIT_BADPARCOUNT;
   if ( np <= 0 ) goto gError;
                                      /* allocate space */
   if ( ( psFF -> pdVal = calloc( np, sizeof(double) ) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
   if ( ( psFF -> pdRefVal = calloc( np, sizeof(double) ) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
   if ( ( psFF -> pdLim = calloc( np, sizeof(double)*2 ) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
   if ( ( psFF -> pdConfLim = calloc( np, sizeof(double)*2 ) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
   if ( ( psFF -> plFlag = calloc( np, sizeof(unsigned long) ) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
/*
   if ( ( psFF -> plRefNum = calloc( np, sizeof(long) ) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
   if ( ( psFF -> plRefType = calloc( np, sizeof(long) ) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
*/
                                      /* remember size */
   psFF -> np = np;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup */
   if ( ( iRC < 0 ) && psFF ) fitFctFree( psFF, lFlag );

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctFree( struct SFITFCT *psFF,
                unsigned long lFlag
              ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctFree)", *pMsg = pDbgMsg;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFF == NULL ) goto gError;
                                      /* release user function */
   if ( (psFF -> lType == FITFCT_USR) && (psFF->lFlagNoDup & FITFCT_UNLOAD) ) {
      if ( psFF -> eLoad ) {
         fitusr( psFF, NULL, NULL, NULL, NULL, psFF->np, 0, FITFCT_CLEANUP);
      }
      if ( psFF->hDyn ) dynunload( psFF->hDyn );
   }
                                      /* release binary tree function */
   if ( psFF -> lType == FITFCT_BTREE ) {
      if ( psFF -> pvBT ) {
         fitbtree( psFF, NULL, NULL, NULL, NULL, psFF->np, 0, FITFCT_CLEANUP);
      }
      btreeprc( psFF->pvBT, NULL, 0, BTREE_CLEANUP );
   }
                                      /* */
   if ( psFF->pcText ) free( (char *) psFF->pcText ); psFF->pcText = NULL;
   if ( psFF->pcName ) free( (char *) psFF->pcName ); psFF->pcName = NULL;
                                      /* release parameter related storage */
   if ( psFF -> pdVal ) {
      free( psFF -> pdVal ); psFF -> pdVal = NULL;
   }
   if ( psFF -> pdRefVal ) {
      free( psFF -> pdRefVal ); psFF -> pdRefVal = NULL;
   }
   if ( psFF -> pdLim ) {
      free( psFF -> pdLim ); psFF -> pdLim = NULL;
   }
   if ( psFF -> pdConfLim ) {
      free( psFF -> pdConfLim ); psFF -> pdConfLim = NULL;
   }
   if ( psFF -> plFlag ) {
      free( psFF -> plFlag ); psFF -> plFlag = NULL;
   }
/*
   if ( psFF -> plRefNum ) {
      free( psFF -> plRefNum ); psFF -> plRefNum = NULL;
   }
   if ( psFF -> plRefType ) {
      free( psFF -> plRefType ); psFF -> plRefType = NULL;
   }
*/
   psFF->np = 0;

   fitParDerivFree( &psFF->sFPD, lFlag );

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctInit( struct SFITFCT *psFF,
                unsigned long lFlag
              ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctInit)", *pMsg = pDbgMsg;
   int ip;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFF == NULL ) goto gError;

   psFF -> lStatus = 0;
   psFF -> lType = 0;
   psFF -> pcText = NULL;
   psFF -> pcName = NULL;
   psFF -> eCall = NULL;
   psFF -> eLoad = NULL;
   psFF -> pvBT  = NULL;
   psFF -> pvBTI = NULL;

   if ( psFF->np > 0 ) {
      if ( psFF->pdVal )
         memset( psFF->pdVal, 0, psFF->np * sizeof(double) );
      if ( psFF->pdRefVal )
         memset( psFF->pdRefVal, 0, psFF->np * sizeof(double) );
      if ( psFF->plFlag )
         memset( psFF->plFlag, 0, psFF->np * sizeof(unsigned long) );
/*
      if ( psFF->plRefNum )
         memset( psFF->plRefNum, 0, psFF->np * sizeof(long) );
      if ( psFF->plRefType )
         memset( psFF->plRefType, 0, psFF->np * sizeof(long) );
*/
      if ( psFF->pdLim )
         for ( ip=0; ip<psFF->np; ip++ ) {
            psFF->pdLim[ ip*2 + 0 ] = -DBL_MAX;
            psFF->pdLim[ ip*2 + 1 ] =  DBL_MAX;
         }
      if ( psFF->pdConfLim )
         for ( ip=0; ip<psFF->np; ip++ ) {
            psFF->pdConfLim[ ip*2 + 0 ] = 0E0;
            psFF->pdConfLim[ ip*2 + 1 ] = 0E0;
         }
   }

   fitParDerivInit( &psFF->sFPD, lFlag );

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctInspect( struct SFITFCT *psFF,
                   unsigned long lFlag
                 ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctInspect)", *pMsg = pDbgMsg;
   int ip;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFF == NULL ) goto gError;

   fprintf( stderr, "<D> SFITFCT=%X\n", psFF );
   fprintf( stderr, "<D> status=%X nodup=%X type=%ld\n",
                    psFF->lStatus, psFF->lFlagNoDup, psFF->lType );
   if ( psFF->pcText )
      fprintf( stderr, "<D> text=%s\n", psFF->pcText );
   if ( psFF->pcName )
      fprintf( stderr, "<D> name=%s\n", psFF->pcName );
   fprintf( stderr, "<D> hdyn=%X\n", psFF->hDyn );
   fprintf( stderr, "<D> call=%X load=%X\n", psFF->eCall, psFF->eLoad );
   fprintf( stderr, "<D> BT=%X BTI=%X\n", psFF->pvBT, psFF->pvBTI );

   fprintf( stderr, "<D> np=%d\n", psFF->np );

   for ( ip=0; ip<psFF->np; ip++ ) {
      fprintf( stderr, "<D> %G %G %G %G %G %G %X\n",
                       psFF->pdVal[ip], psFF->pdRefVal[ip],
                       psFF->pdLim[ip*2+0], psFF->pdLim[ip*2+1],
                       psFF->pdConfLim[ip*2+0], psFF->pdConfLim[ip*2+1],
                       psFF->plFlag[ip]
             );
   }

   fitParDerivInspect( &psFF->sFPD, lFlag );

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
