#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "m.h"
#include "btree.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitfctblk.c,v 1.4 1999/08/02 09:21:14 kraemer Exp $"};

/*
   --- Functions to handle function block related structures

       fitFctBlkAlloc()   : allocate
       fitFctBlkFree()    : free
       fitFctBlkInit()    : initialize
       fitFctBlkInspect() : inspect output
       fitFctBlkDup()     : duplicate a function block

       fitFctBlkFctPut()  : add/copy a single function

       fitFctBlkParGetByNum() : retrieves parameter values
       fitFctBlkParPutByNum() : stores    parameter values

       fitFctBlkParBlkGet() : retrieves a block of parameter values
       fitFctBlkParBlkPut() : stores    a block of parameter values

       fitFctBlkPeak2Par()  : converts peak numbers to parameter number offs
       fitFctBlkPar2Peak()  : converts parameter number to peak number 
       fitFctBlkPeaks()     : collect peak functions
       fitFctBlkByNum()     : collect functions by number

       fitFctBlkFctDelByNum(): delete function by number
       fitFctBlkCovar()     : compute covariance from curvature matrix
       fitFctBlkCovarStr()  : covariance matrix -> char string

       fitFctBlkCorrCopy()  : copies correlation arrays

       fitFctBlkParStrByNum(): parameter values -> char string

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitfctblk.c,v $
   Revision 1.4  1999/08/02 09:21:14  kraemer
   *** empty log message ***

   Revision 1.2  1997/12/15 18:59:59  kraemer
   include "no data errors", fit statistics

   Revision 1.1.1.1  1997/07/10 18:00:05  kraemer
   revision for cvs

*/


int fitFctBlkDup( struct SFITFCTBLK *psFFBFr,
                  struct SFITFCTBLK *psFFBTo,
                  unsigned long lFlag
                ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkDup)", *pMsg = pDbgMsg;

   int jf;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFBTo == NULL ) goto gError;
   memset( psFFBTo, 0, sizeof(psFFBTo) );
   if ( psFFBFr == NULL ) goto gError;
                                      /* allocate new block */
   if ( iRC = fitFctBlkAlloc( psFFBFr->nf, psFFBFr->np, psFFBTo, lFlag ) )
      goto gError;
                                      /* copy stuff */
   psFFBTo -> lStatus = psFFBFr -> lStatus;
   for ( jf=0; jf<psFFBTo->nf; jf++ ) {
      if ( iRC = fitFctDup( psFFBFr->psFF+jf, psFFBTo->psFF+jf, lFlag ) ) goto gError;
   }
                                      /*  */
   memcpy( psFFBTo->pdBeta, psFFBFr->pdBeta, psFFBTo->np * sizeof(double) );
   memcpy( psFFBTo->pdDiag, psFFBFr->pdDiag, psFFBTo->np * sizeof(double) );
   memcpy( psFFBTo->pdAlpha, psFFBFr->pdAlpha, psFFBTo->np * psFFBTo->np * sizeof(double) );
   memcpy( psFFBTo->pdCovar, psFFBFr->pdCovar, psFFBTo->np * psFFBTo->np * sizeof(double) );
   memcpy( psFFBTo->plRow,    psFFBFr->plRow,    psFFBTo->np * sizeof(long) );
   memcpy( psFFBTo->plCol,    psFFBFr->plCol,    psFFBTo->np * sizeof(long) );
   memcpy( psFFBTo->plRefNum, psFFBFr->plRefNum, psFFBTo->np * sizeof(long) );
   memcpy( psFFBTo->plRefTyp, psFFBFr->plRefTyp, psFFBTo->np * sizeof(long) );
   psFFBTo->sFS = psFFBFr->sFS;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
   if ( (iRC < 0) && psFFBTo ) {
      fitFctBlkFree( psFFBTo, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}

/*
   =========================================================================
*/
int fitFctBlkAlloc( int nf, int np,
                    struct SFITFCTBLK *psFFB,
                    unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkAlloc)", *pMsg = pDbgMsg;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   iRC = ERRFIT_BADFUNCTIONCOUNT;
   if ( nf <= 0 ) goto gError;
   iRC = ERRFIT_BADPARCOUNT;
   if ( np <= 0 ) goto gError;
                                      /* allocate space for functions */
   if ( ( psFFB -> psFF= calloc( np, sizeof(struct SFITFCT) ) ) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
                                      /* allocate space for correlations */
   iRC = ERRFIT_SYS;
   if ( (psFFB -> pdBeta  = calloc(np,sizeof(double)) ) == NULL ) goto gError;
   if ( (psFFB -> pdDiag  = calloc(np,sizeof(double)) ) == NULL ) goto gError;
   if ( (psFFB -> pdAlpha = calloc(np,np*sizeof(double)) ) == NULL ) goto gError;
   if ( (psFFB -> pdCovar = calloc(np,np*sizeof(double)) ) == NULL ) goto gError;
   if ( (psFFB -> plRow    = calloc(np,sizeof(long)) ) == NULL ) goto gError;
   if ( (psFFB -> plCol    = calloc(np,sizeof(long)) ) == NULL ) goto gError;
   if ( (psFFB -> plRefNum = calloc(np,sizeof(long)) ) == NULL ) goto gError;
   if ( (psFFB -> plRefTyp = calloc(np,sizeof(long)) ) == NULL ) goto gError;
                                      /* remember parameter and function # */
   psFFB -> nf = nf;
   psFFB -> np = np;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
   if ( (iRC < 0) && psFFB ) {
      fitFctBlkFree( psFFB, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}

/*
   =========================================================================
*/
int fitFctBlkFree( struct SFITFCTBLK *psFFB,
                   unsigned long lFlag
                 ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkFree)", *pMsg = pDbgMsg;
   int jf; struct SFITFCT *psFF;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
                                      /* release single functions */
   if ( psFF = psFFB->psFF ) {
      for ( jf=0; jf<psFFB->nf; jf++ ) {
         fitFctFree( psFF, lFlag ); psFF++;
      }
   }
                                      /* release correlation arrays */
   if ( psFFB -> pdBeta ) free( psFFB -> pdBeta ); psFFB -> pdBeta = NULL;
   if ( psFFB -> pdDiag ) free( psFFB -> pdDiag ); psFFB -> pdDiag = NULL;
   if ( psFFB -> pdAlpha) free( psFFB -> pdAlpha); psFFB -> pdAlpha= NULL;
   if ( psFFB -> pdCovar) free( psFFB -> pdCovar); psFFB -> pdCovar= NULL;
   if ( psFFB -> plRow  ) free( psFFB -> plRow  ); psFFB -> plRow  = NULL;
   if ( psFFB -> plCol  ) free( psFFB -> plCol  ); psFFB -> plCol  = NULL;
   if ( psFFB -> plRefTyp ) free( psFFB -> plRefTyp ); psFFB -> plRefTyp = NULL;
   if ( psFFB -> plRefNum ) free( psFFB -> plRefNum ); psFFB -> plRefNum = NULL;

   psFFB->np = 0;
   psFFB->nf = 0;
   psFFB->lStatus &= ~FITFCTBLK_KEEP;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctBlkInit( struct SFITFCTBLK *psFFB,
                   unsigned long lFlag
                 ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkInit)", *pMsg = pDbgMsg;
   int np, jf; struct SFITFCT *psFF;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
                                      /* initialize single functions */
   if ( psFF=psFFB->psFF ) {
      for ( jf=0; jf<psFFB->nf; jf++ ) {
         fitFctInit( psFF, lFlag ); psFF++;
      }
   }

   if ( ( np = psFFB->np ) > 0 ) {
      if ( psFFB->pdBeta ) memset( psFFB->pdBeta,  0,    np*sizeof(double) );
      if ( psFFB->pdDiag ) memset( psFFB->pdDiag,  0,    np*sizeof(double) );
      if ( psFFB->pdAlpha) memset( psFFB->pdAlpha, 0, np*np*sizeof(double) );
      if ( psFFB->pdCovar) memset( psFFB->pdCovar, 0, np*np*sizeof(double) );
      if ( psFFB->plCol  ) memset( psFFB->plCol, 0, np*sizeof(long) );
      if ( psFFB->plRow  ) memset( psFFB->plRow, 0, np*sizeof(long) );
      if ( psFFB->plRefTyp ) memset( psFFB->plRefTyp, 0, np*sizeof(long) );
      if ( psFFB->plRefNum ) memset( psFFB->plRefNum, 0, np*sizeof(long) );
   }

   psFFB->lStatus = 0;
   fitStatInit( &psFFB->sFS, lFlag );

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctBlkInspect( struct SFITFCTBLK *psFFB,
                      unsigned long lFlag
                    ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkInspect)", *pMsg = pDbgMsg;
   int jf; struct SFITFCT *psFF;
   int ip1, ip2;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;

   fprintf( stderr, "<D> SFITFCTBLK=%X\n", psFFB );
   fprintf( stderr, "<D> status=%X\n", psFFB->lStatus );
   fprintf( stderr, "<D> nf=%d np=%d lambda=%G\n",
                    psFFB->nf, psFFB->np, psFFB->sFS.dLambda );
   fprintf( stderr, "<D> beta=%X diag=%X alpha=%X covar=%X\n",
                         psFFB->pdBeta, psFFB->pdDiag,
                         psFFB->pdAlpha, psFFB->pdCovar );
   fitStatInspect( &psFFB->sFS, lFlag ); 
   fprintf( stderr, "<D> psFF=%X\n", psFFB->psFF );
   if ( psFF=psFFB->psFF ) {
      for ( jf=0; jf<psFFB->nf; jf++ ) {
         fitFctInspect( psFF, lFlag ); psFF++;
      }
   }

   if ( psFFB->pdBeta && psFFB->pdDiag )
      for ( ip1=0; ip1<psFFB->np; ip1++ ) {
         fprintf( stderr, "<D> beta=%G diag=%G reftyp=%d refnum=%d\n",
                          psFFB->pdBeta[ip1], psFFB->pdDiag[ip1],
                          psFFB->plRefTyp[ip1], psFFB->plRefNum[ip1]
                );
      }
   if ( psFFB->pdAlpha )
      for ( ip1=0; ip1<psFFB->np; ip1++ ) {
         fprintf( stderr, "<D> alpha=" );
         for ( ip2=0; ip2<psFFB->np; ip2++ ) {
            fprintf( stderr, "%G ", psFFB->pdAlpha[ip1*psFFB->np+ip2] );
         }
         fprintf( stderr, "\n" );
      }
   if ( psFFB->pdCovar )
      for ( ip1=0; ip1<psFFB->np; ip1++ ) {
         fprintf( stderr, "<D> covar=" );
         for ( ip2=0; ip2<psFFB->np; ip2++ ) {
            fprintf( stderr, "%G ", psFFB->pdCovar[ip1*psFFB->np+ip2] );
         }
         fprintf( stderr, "\n" );
      }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}

/*
   =========================================================================
*/
int fitFctBlkFctPut( struct SFITFCTBLK *psFFB,
                     struct SFITFCT *psFF,
                     unsigned long lFlag
                   ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkFctPut)", *pMsg = pDbgMsg;
   int jf;
   long lvIndex[1][2] = {0,0};
   struct SFITFCTBLK sFFB;
   unsigned long lFlagDup = 0;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sFFB, 0, sizeof(sFFB) ); 
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   sFFB.sFS = psFFB->sFS;
                                      /* append ? */
   if ( psFFB -> lStatus & FITFCTBLK_KEEP ) {
                                      /* space for one more function */
      if ( (iRC = fitFctBlkAlloc( psFFB->nf+1, psFFB->np + psFF -> np,
                                  &sFFB, lFlag ) != 0) ) goto gError;
                                      /* copy old */
      for ( jf=0; jf<psFFB->nf; jf++ ) {
         sFFB.psFF[jf] = psFFB->psFF[jf];
      }
/*
      sFFB.nf = psFFB -> nf + 1;
      sFFB.np = np = psFFB -> np + psFF -> np;
*/
      sFFB.lStatus = psFFB -> lStatus;
                                      /* append new function */
      if ( lFlag & FITFCTBLKFCTPUT_DUPENABLE )
         lFlagDup |= FITFCTDUP_DUPENABLE;
      if ( ( iRC = fitFctDup( psFF, sFFB.psFF + psFFB->nf, lFlag | lFlagDup ) ) != 0 )
         goto gError;
                                      /* copy old correlations */
      if ( (lvIndex[0][1] = psFFB->np-1) >= 0 ) {
         if ( ( iRC = fitFctBlkCorrCopy(psFFB,lvIndex,1,&sFFB,lFlag) ) != 0 )
            goto gError;
      }
                                      /* free old allocations */
      free( psFFB->psFF ); psFFB->psFF = NULL;
      fitFctBlkFree( psFFB, lFlag );
   } else {                           /* replace ? */
                                      /* space for new function */
      if ( (iRC = fitFctBlkAlloc( 1, psFF->np, &sFFB, lFlag ) != 0) )
         goto gError;
                                      /* duplicate function */
      if ( ( iRC = fitFctDup(psFF,sFFB.psFF,lFlag) ) != 0 ) goto gError;
/*
      sFFB.nf = 1;
      sFFB.np = np = psFF->np;
*/
                                      /* release old, if any */
      fitFctBlkFree( psFFB, lFlag );
   }
   sFFB.sFS.dLambda = 1E-3;
                                      /* delete old, link new */
   *psFFB = sFFB; memset( &sFFB, 0, sizeof(sFFB) );

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup in case of error */
   if ( iRC < 0 ) {
      fitFctBlkFree( &sFFB, lFlag );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}

/*
   =========================================================================
*/
int fitFctBlkParGetByNum( struct SFITFCTBLK *psFFB,
                          long lPar,  /* parameter number (total,1..npar)*/
                          struct SFITPAR *psFP,
                          unsigned long lFlag ) {

   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkParGetByNum)", *pMsg = pDbgMsg;
   int jf;
   struct SFITFCT *psFF;
   long lParB;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   if ( psFP  == NULL ) goto gError;
   iRC = ERRFIT_BADPARNUM;
   if ( lPar <= 0 ) goto gError;
   if ( lPar > psFFB -> np ) goto gError;
   lParB = lPar-1;
                                      /* loop over functions to locate par*/
   psFF = psFFB -> psFF;
   for ( jf=0; jf<psFFB->nf; jf++ ) {
      if ( lPar <= psFF->np ) {
         lPar--;                      /* adjust index */
                                      /* return parameter values et al */
         psFP -> dVal     = psFF -> pdVal[lPar];
         psFP -> dRefVal  = psFF -> pdRefVal[lPar];
         psFP -> dLim[0]  = psFF -> pdLim[lPar*2+0];
         psFP -> dLim[1]  = psFF -> pdLim[lPar*2+1];
         psFP -> lRefTyp  = psFFB -> plRefTyp[lParB];
         psFP -> lRefNum  = psFFB -> plRefNum[lParB];
         psFP -> lFlag    = psFF -> plFlag  [lPar];
         psFP -> dConfLim[0] = psFP -> dConfLim[1] =
         psFP -> dErr[0] = psFP -> dErr[1] = psFP -> dErrSym = 0E0;
         if ( psFFB -> pdCovar != NULL ) {
            psFP -> dErr[0] = psFP -> dErr[1] = psFP -> dErrSym = 
            sqrt( fabs( psFFB->pdCovar[ lParB * psFFB->np + lParB ] ) );
         }
                                      /* try asymmetric errors */
         if ( psFP -> lFlag & FITPAR_EON ) {
            psFP -> dErr[0] = psFP -> dConfLim[0] = psFF -> pdConfLim[lPar*2+0];
            psFP -> dErr[1] = psFP -> dConfLim[1] = psFF -> pdConfLim[lPar*2+1];
         }

                                      /* apply correction for data w/o errors */
         if ( psFFB->sFS.lStatus & FITSTAT_NODATAERR ) {
            if ( psFFB->sFS.nDegFree > 0 ) { 
               psFP -> dErr[0] *= sqrt( psFFB->sFS.dChisq / psFFB->sFS.nDegFree ); 
               psFP -> dErr[1] *= sqrt( psFFB->sFS.dChisq / psFFB->sFS.nDegFree );
            }
         }
         iRC = 0;                     /* !!! set return code = OK here */
         break;
      }
      lPar -= psFF->np; psFF++;       /* next function */
   }
                                      /* */
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctBlkParPutByNum( struct SFITFCTBLK *psFFB,
                          long lPar,
                          struct SFITPAR *psFP,
                          unsigned long lFlag ) {

   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkParPutByNum)", *pMsg = pDbgMsg;
   int jf; struct SFITFCT *psFF;
   long lParB;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   if ( psFP  == NULL ) goto gError;
   iRC = ERRFIT_BADPARNUM;
   if ( lPar <= 0 ) goto gError;
   if ( lPar > psFFB -> np ) goto gError;
   lParB = lPar-1;                    /* parameter index */

   if ( psFP -> lRefTyp > 0 ) {
      if ( psFP -> lRefNum < 0 ) goto gError;
      if ( psFP -> lRefNum >= psFFB -> np ) goto gError;
      iRC = ERRFIT_BADPARREF;
      if ( psFP -> lRefNum == lParB ) goto gError;
   }
                                      /* loop over functions to locate par*/
   psFF = psFFB -> psFF;
   for ( jf=0; jf<psFFB->nf; jf++ ) {
      if ( lPar <= psFF->np ) {
         lPar--;                      /* adjust index */
                                      /* return parameter values et al */
         psFF -> pdVal    [lPar] = psFP -> dVal;
         psFF -> pdRefVal [lPar] = psFP -> dRefVal;
         psFF -> pdLim[lPar*2+0] = psFP -> dLim[0];
         psFF -> pdLim[lPar*2+1] = psFP -> dLim[1];
         psFF -> pdConfLim[lPar*2+0] = psFP -> dConfLim[0];
         psFF -> pdConfLim[lPar*2+1] = psFP -> dConfLim[1];
         if ( (psFFB -> plRefTyp[lParB] = psFP -> lRefTyp) > 0 ) 
            psFFB -> plRefNum[lParB] = psFP -> lRefNum;
         else 
            psFFB -> plRefNum[lParB] = 0;
         psFF -> plFlag  [lPar] = psFP -> lFlag;
         iRC = 0;                     /* !!! set return code = OK here */
         break;
      }
      lPar -= psFF->np; psFF++;       /* next function */
   }
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}

/*
   =========================================================================
*/
int fitFctBlkParBlkGet(   struct SFITFCTBLK *psFFB,
                          struct SFITPARBLK *psFPB,
                          unsigned long lFlag ) {

   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkParBlkGet)", *pMsg = pDbgMsg;
   int jf, ip, np=0;
   struct SFITFCT *psFF;
   struct SFITPAR *psFP;
   long lParB=0;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   if ( psFPB == NULL ) goto gError;
   iRC = ERRFIT_BADPARCOUNT;
   if ( (np = psFFB->np) <= 0 ) goto gError;
                                      /* allocate a parameter block */
   if ( ( iRC = fitParBlkAlloc(np,psFPB,lFlag) ) != 0 ) goto gError;
                                      /* function loop */
   psFF = psFFB -> psFF;
   psFP = psFPB -> psFP;
   np   = psFPB -> np;
   for ( jf=0; jf<psFFB->nf; jf++ ) {
                                      /* for security: check consistence */
      if ( (np -= psFF->np) < 0 ) {
         iRC = ERRFIT_PARCOUNTINCONSISTENT;
         goto gError;
      }
      for ( ip=0; ip<psFF->np; ip++ ) {
         psFP -> dVal     = psFF -> pdVal    [ip];
         psFP -> dRefVal  = psFF -> pdRefVal [ip];
         psFP -> dLim[0]  = psFF -> pdLim    [ip*2+0];
         psFP -> dLim[1]  = psFF -> pdLim    [ip*2+1];
         psFP -> dConfLim[0] = psFF -> pdConfLim[ip*2+0];
         psFP -> dConfLim[1] = psFF -> pdConfLim[ip*2+1];
         psFP -> lRefTyp  = psFFB -> plRefTyp[lParB];
         psFP -> lRefNum  = psFFB -> plRefNum[lParB];
         psFP -> lFlag    = psFF -> plFlag  [ip];
         psFP++;
         lParB++;
      }
      psFF++;                         /* next function */
   }
                                      /* */
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup in case of error */
   if ( (iRC < 0) && psFPB ) {
      fitParBlkFree( psFPB, lFlag );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctBlkParBlkPut(   struct SFITFCTBLK *psFFB,
                          struct SFITPARBLK *psFPB,
                          unsigned long lFlag ) {

   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkParBlkPut)", *pMsg = pDbgMsg;
   int jf, ip, np=0;
   struct SFITFCT *psFF;
   struct SFITPAR *psFP;
   long lParB = 0;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   if ( psFPB == NULL ) goto gError;
   iRC = ERRFIT_PARCOUNTINCONSISTENT;
   if ( psFFB->np != psFPB->np ) goto gError;
   iRC = ERRFIT_BADPARCOUNT;
   if ( (np = psFPB->np) <= 0 ) goto gError;
                                      /* check valid addresses */
   iRC = ERRFIT_BADADDR;
   if ( (psFF = psFFB -> psFF) == NULL ) goto gError;
   if ( (psFP = psFPB -> psFP) == NULL ) goto gError;
                                      /* loop over functions */
   for ( jf=0; jf<psFFB->nf; jf++ ) {
                                      /* for security: check consistence */
      if ( (np -= psFF->np) < 0 ) {
         iRC = ERRFIT_PARCOUNTINCONSISTENT;
         goto gError;
      }
      for ( ip=0; ip<psFF->np; ip++ ) {
         psFF -> pdVal    [ip]     = psFP -> dVal;
         psFF -> pdRefVal [ip]     = psFP -> dRefVal;
         psFF -> pdLim    [ip*2+0] = psFP -> dLim[0];
         psFF -> pdLim    [ip*2+1] = psFP -> dLim[1];
         psFF -> pdConfLim[ip*2+0] = psFP -> dConfLim[0];
         psFF -> pdConfLim[ip*2+1] = psFP -> dConfLim[1];
         psFFB-> plRefTyp[lParB]   = psFP -> lRefTyp;
         psFFB-> plRefNum[lParB]   = psFP -> lRefNum;
         psFF -> plFlag  [ip]     = psFP -> lFlag;
         psFP++;                      /* next parameter */
         lParB++;
      }
      psFF++;                         /* next function */
   }
                                      /* */
   iRC = 0;

gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctBlkPeak2Par( struct SFITFCTBLK *psFFB,
                       long lvPeak[],
                       int nPeak,
                       long lvPar[] ,
                       unsigned long lFlag ) {

   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkPeak2Par)", *pMsg = pDbgMsg;
   int jf, ip, nParPerPeak;
   long lPeakMin, lPeakMax, lPar;
   struct SFITFCT *psFF;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   if ( lvPeak == NULL ) goto gError;
   if ( lvPar == NULL ) goto gError;

   iRC = ERRFIT_BADPEAKCOUNT;
   if ( nPeak <= 0 ) goto gError;
                                      /* check valid addresses */
   iRC = ERRFIT_BADADDR;
   if ( psFFB -> psFF == NULL ) goto gError;
                                      /* loop over requested peaks */
   for ( ip=0; ip<nPeak; ip++ ) {
                                      /* loop over functions */
      psFF = psFFB -> psFF;
      lPeakMin = 1; lPeakMax = 0;     /* reset peak numbers */
      lPar = lvPar[ip] = 0;           /* reset par numbers */
      for ( jf=0; jf<psFFB->nf; jf++ ) {
                                      /* check for peak functions */
         nParPerPeak = FITFCT_PARPERPEAK;
         switch( psFF -> lType ) {
         case FITFCT_GAUSSBOX:
         case FITFCT_LORENTZBOX:
            nParPerPeak = FITFCT_PARPERBOX;
         case FITFCT_LORENTZ:
         case FITFCT_GAUSS:
                                      /* peak number range this function */
            lPeakMax = lPeakMin + psFF->np / nParPerPeak - 1;
            if ( ( lvPeak[ip] >= lPeakMin ) &&
                 ( lvPeak[ip] <= lPeakMax )
               ) {
               lvPar[ip] = lPar + (lvPeak[ip]-lPeakMin) * nParPerPeak + 1;
            }
            lPeakMin = lPeakMax + 1;
         break;
         default:
         break;
         }
         if ( lvPar[ip] > 0 ) break;  /* if peak found: leave */
         lPar += psFF->np;            /* running parameter number */
         psFF++;                      /* next function */
      }
      if ( lvPar[ip] <= 0 ) {
         iRC = ERRFIT_BADPEAKNUM; goto gError;
      }
   }
                                      /* */
   iRC = 0;

gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctBlkPar2Peak( struct SFITFCTBLK *psFFB,
                       long lvPar[],
                       int nPar,
                       long lvPeak[] ,
                       unsigned long lFlag ) {

   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkPar2Peak)", *pMsg = pDbgMsg;

   struct SFITFCT *psFF;
   int jf, ip, nParPerPeak;
   long lPar, lPeakMin, lPeakMax;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   if ( lvPeak == NULL ) goto gError;
   if ( lvPar == NULL ) goto gError;

   iRC = ERRFIT_BADPARCOUNT;
   if ( nPar <= 0 ) goto gError;
                                      /* check valid addresses */
   iRC = ERRFIT_BADADDR;
   if ( psFFB -> psFF == NULL ) goto gError;
                                      /* loop over requested parameters */
   for ( ip=0; ip<nPar; ip++ ) {
                                      /* loop over functions */
      psFF = psFFB -> psFF;
      lPeakMin = 1; lPeakMax = 0;     /* reset peak numbers */
      lvPeak[ip] = 0;
      lPar = lvPar[ip];
      iRC = ERRFIT_BADPARNUM;
      for ( jf=0; jf<psFFB->nf; jf++ ) {
                                      /* determine peak range */
         nParPerPeak = FITFCT_PARPERPEAK;
         switch( psFF -> lType ) {
         case FITFCT_GAUSSBOX:
         case FITFCT_LORENTZBOX:
            nParPerPeak = FITFCT_PARPERBOX;
         case FITFCT_LORENTZ:
         case FITFCT_GAUSS:
                                      /* peak number range this function */
            lPeakMin = lPeakMax + 1;
            lPeakMax = lPeakMin + psFF->np / nParPerPeak - 1;
         break;
         default:
         break;
         }

         if ( lPar <= psFF->np ) {    /* parameter within function ? */
            lPar--;                   /* adjust index, parameter offs */
            iRC = 0; break;
         }
         lPar -= psFF->np; psFF++;    /* next function */
      }
      if ( iRC < 0 ) goto gError;     /* parameter not found */

                                      /* if found, check if peak function */
      nParPerPeak = FITFCT_PARPERPEAK;
      switch( psFF -> lType ) {
      case FITFCT_GAUSSBOX:
      case FITFCT_LORENTZBOX:
         nParPerPeak = FITFCT_PARPERBOX;
      case FITFCT_LORENTZ:
      case FITFCT_GAUSS:
         lvPeak[ip] = lPar / nParPerPeak + lPeakMin;
      break;
      default:
/* ??        iRC = ERRFIT_...; goto gError; */
      break;
      }
   }
                                      /* */
   iRC = 0;

gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctBlkPeaks( struct SFITFCTBLK *psFFBFr,
                    struct SFITFCTBLK *psFFBTo,
                    unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkPeaks)", *pMsg = pDbgMsg;
   int jf, nf, np, ipf;
   struct SFITFCT *psFFFr, *psFFTo, sFF;

   long *lvSubIndex = NULL; long lIndex; int kf;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFBFr == NULL ) goto gError;
   if ( psFFBTo == NULL ) goto gError;
   memset( psFFBTo, 0, sizeof(*psFFBTo) );
                                      /* */
   psFFFr = psFFBFr -> psFF; nf = 0; np = 0;
   for ( jf=0; jf<psFFBFr->nf; jf++ ) {
                                      /* check for peak functions */
      switch( psFFFr -> lType ) {
      case FITFCT_GAUSSBOX:
      case FITFCT_LORENTZBOX:
      case FITFCT_LORENTZ:
      case FITFCT_GAUSS:
         nf += psFFFr->np/psFFFr->npSpecial; np += psFFFr->np;
      break;
      default:
      break;
      }
      psFFFr++;                       /* next function */
   }
                                      /* any peak functions ? */
   if ( nf <= 0 ) {
      iRC = ERRFIT_NOPEAKFUNCTION; goto gError;
   }
                                      /* space for peak function */
   if ( ( iRC = fitFctBlkAlloc( nf, np, psFFBTo, lFlag ) ) != 0 ) goto gError;
   psFFBTo -> lStatus = 0;
                                      /* space for index range */
   if ( (lvSubIndex = (long *) calloc( sizeof(long), nf*2 )) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
                                      /* loop to collect peak functions */
   lIndex = 0; kf = 0;
   psFFTo = psFFBTo -> psFF;
   psFFFr = psFFBFr -> psFF;
   for ( jf=0; jf<psFFBFr->nf; jf++ ) {
                                      /* check for peak functions */
      switch( psFFFr -> lType ) {
      case FITFCT_GAUSSBOX:
      case FITFCT_LORENTZBOX:
      case FITFCT_LORENTZ:
      case FITFCT_GAUSS:
                                      /* reference copy */
         sFF = *psFFFr;
                                      /* fake a single peak */;
         sFF.np = psFFFr -> npSpecial;
         memset( &sFF.sFPD, 0, sizeof(sFF.sFPD) );                             
         for( ipf=0; ipf<psFFFr -> np; ipf += psFFFr -> npSpecial ) {   
            if ( ( iRC = fitFctDup( &sFF, psFFTo, lFlag ) ) ) goto gError;
//fitFctInspect( psFFTo, lFlag );            
            lvSubIndex[kf*2+0] = lIndex;
            lvSubIndex[kf*2+1] = lIndex + psFFFr -> npSpecial - 1;
            kf++;
            psFFTo++;
            lIndex += psFFFr -> npSpecial;
                                      /* fake parameter addresses */
            sFF.pdVal += psFFFr -> npSpecial;
            sFF.pdRefVal += psFFFr -> npSpecial;
            sFF.pdLim += psFFFr -> npSpecial * 2;
            sFF.pdConfLim += psFFFr -> npSpecial * 2;
            sFF.plFlag += psFFFr -> npSpecial;
         }   
      break;
      default:
      break;
      }
      lIndex += psFFFr -> np;
      psFFFr++;                       /* next function */
   }
                                      /* correlations */
/*
fprintf( stderr, "<D> %d %d %d %d\n",
                 lvSubIndex[0], lvSubIndex[1],
                 lvSubIndex[2], lvSubIndex[3] );
*/
   if ( ( iRC = fitFctBlkCorrCopy( psFFBFr, (long (* )[2]) lvSubIndex, nf, psFFBTo, lFlag) ) != 0 )
      goto gError;
   psFFBTo->sFS = psFFBFr->sFS;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup in case of error */
   if ( (iRC < 0) && psFFBTo ) {
      fitFctBlkFree( psFFBTo, lFlag );
   }
   if ( lvSubIndex ) free( lvSubIndex ); lvSubIndex = NULL;

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}

/*
   =========================================================================
*/
int fitFctBlkByNum( struct SFITFCTBLK *psFFBFr,
                    long lvFct[],
                    int nFct,
                    struct SFITFCTBLK *psFFBTo,
                    unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkByNum)", *pMsg = pDbgMsg;
   int jf, kf, nf, np;
   struct SFITFCT *psFFFr, *psFFTo;

   long *lvSubIndex = NULL; long lIndex;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFBFr == NULL ) goto gError;
   if ( psFFBTo == NULL ) goto gError;
   if ( lvFct == NULL ) goto gError;
   memset( psFFBTo, 0, sizeof(*psFFBTo) );
   iRC = ERRFIT_BADFUNCTIONCOUNT;
   if ( nFct <= 0 ) goto gError;
                                      /* space for index range */
   if ( (lvSubIndex = (long *) calloc( sizeof(long), nFct*2 )) == NULL ) {
      iRC = ERRFIT_SYS; goto gError;
   }
                                      /* check function numbers */
   iRC = ERRFIT_BADFUNCTIONNUM;
   nf = np = 0;
   for ( jf=0; jf<nFct; jf++ ) {
      if ( lvFct[jf] <= 0 ) goto gError;
      if ( lvFct[jf] > psFFBFr->nf ) goto gError;
      psFFFr = psFFBFr -> psFF + lvFct[jf] - 1;
      np += psFFFr -> np; 
      nf++;
   }
                                      /* any functions ? */
   if ( nf <= 0 ) {
      iRC = ERRFIT_NOFUNCTION; goto gError;
   }
                                      /* determine index ranges */
   for ( jf=0; jf<nFct; jf++ ) {
      lIndex = 0; psFFFr = psFFBFr -> psFF;
      for ( kf=0; kf<psFFBFr->nf; kf++ ) {
         if ( kf == (lvFct[jf]-1) ) {
            lvSubIndex[jf*2+0] = lIndex;
            lvSubIndex[jf*2+1] = lIndex + psFFFr -> np - 1;
         }
         lIndex += psFFFr -> np;
         psFFFr++;
      }
   }
                                      /* space for functions */
   if ( ( iRC = fitFctBlkAlloc( nf, np, psFFBTo, lFlag ) ) != 0 ) goto gError;
   psFFBTo -> lStatus = 0;
                                      /* loop to collect functions */
   psFFTo = psFFBTo -> psFF;
   for ( jf=0; jf<nFct; jf++ ) {
      if ( lvFct[jf] <= 0 ) goto gError;
      if ( lvFct[jf] > psFFBFr->nf ) goto gError;
      psFFFr = psFFBFr -> psFF + lvFct[jf] - 1;
                                      /* duplicate function */
      if ( ( iRC = fitFctDup( psFFFr, psFFTo, lFlag ) ) ) goto gError;
      psFFTo++;
   }
                                      /* correlations */
/*
fprintf( stderr, "<D> %d %d %d %d\n",
                 lvSubIndex[0], lvSubIndex[1],
                 lvSubIndex[2], lvSubIndex[3] );
*/
   if ( ( iRC = fitFctBlkCorrCopy( psFFBFr, (long (* )[2]) lvSubIndex, nFct, psFFBTo, lFlag) ) != 0 )
      goto gError;
   psFFBTo->sFS = psFFBFr->sFS;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup in case of error */
   if ( (iRC < 0) && psFFBTo ) {
      fitFctBlkFree( psFFBTo, lFlag );
   }
   if ( lvSubIndex ) free( lvSubIndex ); lvSubIndex = NULL;

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}

/*
   =========================================================================
*/
int fitFctBlkFctDelByNum( struct SFITFCTBLK *psFFB,
                          long lFct,
                          unsigned long lFlag
                        ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkFctDelByNum)", *pMsg = pDbgMsg;

   struct SFITFCTBLK sFFB;
   struct SFITFCT *psFFFr, *psFFTo;
   int jf;
   long lIndex=0, lvSubIndex[2][2]={0,0,0,0}; int nSub=0;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sFFB, 0, sizeof(sFFB) );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   iRC = ERRFIT_BADFUNCTIONNUM;
   if ( lFct <= 0 ) goto gError;
   if ( lFct > psFFB->nf ) goto gError;
   lFct--;                            /* for index */
                                      /* any function left ? */
   if ( (psFFB->nf-1) > 0 ) {
                                      /* space for new functions */
   if ( ( iRC = fitFctBlkAlloc( psFFB->nf-1,
                                psFFB->np - psFFB->psFF[lFct].np,
                                &sFFB, lFlag ) ) != 0 ) goto gError;

   sFFB.lStatus = psFFB->lStatus;

                                      /* determine index range for corr. */
   psFFFr = psFFB->psFF;
   lvSubIndex[0][1] = lvSubIndex[1][1] = psFFB->np-1;
   for ( jf=0; jf<psFFB->nf; jf++ ) {
      if ( jf == lFct-1 ) {
         lvSubIndex[nSub++][1] = lIndex + psFFFr -> np - 1;
      }
      if ( jf == lFct+1 ) {
         lvSubIndex[nSub++][0] = lIndex;
      }
      lIndex += psFFFr -> np;
      psFFFr++;
   }
                                      /* correlations */
/*
fprintf( stderr, "<D> %d %d %d %d\n",
                 lvSubIndex[0][0], lvSubIndex[0][1],
                 lvSubIndex[1][0], lvSubIndex[1][1] );
*/
   if ( ( iRC = fitFctBlkCorrCopy( psFFB, lvSubIndex, nSub, &sFFB, lFlag) ) != 0 )
      goto gError;
                                      /* loop to collect functions */
   psFFFr = psFFB->psFF;
   psFFTo = sFFB.psFF;
   for ( jf=0; jf<psFFB->nf; jf++ ) {
                                      /* delete/copy function */
      if ( lFct == jf ) {
         psFFB -> np -= psFFFr -> np;
         fitFctFree( psFFFr, lFlag );
      } else {
         *psFFTo = *psFFFr; psFFTo++;
      }
      psFFFr++;
   }
                                      /* delete old, link new */
   free( psFFB->psFF ); psFFB->psFF = NULL;

   } /* nf-1 > 0 */

   fitFctBlkFree( psFFB, lFlag ); *psFFB = sFFB; memset(&sFFB,0,sizeof(sFFB));
   if ( psFFB->nf <= 0 ) psFFB->lStatus &= ~FITFCTBLK_KEEP;
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup in case of error */
   if ( (iRC < 0) ) {
      fitFctBlkFree( &sFFB, lFlag );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitFctBlkCovar( struct SFITFCTBLK *psFFB,
                    unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkCovar)", *pMsg = pDbgMsg;
   int ip1, ip2;
   double *pdA, *pdD, *pdC;
   double dd;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   if ( ( pdA = psFFB->pdAlpha ) == NULL ) goto gError;
   if ( ( pdC = psFFB->pdCovar ) == NULL ) goto gError;
   if ( ( pdD = psFFB->pdDiag  ) == NULL ) goto gError;
   if ( ( psFFB->plCol ) == NULL ) goto gError;
   if ( ( psFFB->plRow ) == NULL ) goto gError;

   for ( ip1=0; ip1<psFFB->np; ip1++ ) {
                                      /* find max in row */
      dd = 0E0;
      for ( ip2=0; ip2<psFFB->np; ip2++ ) {
         if ( fabs( pdA[ip2] ) > dd ) dd = fabs( pdA[ip2] );
      }
                                      /* if all zeroes: diagonal=1 for inv. */
      pdD[ip1] = 0E0;
      if ( dd == 0E0 ) { pdD[ip1] = -1E0; pdA[ip1] = 1E0; }
                                      /* next row */
      pdA += psFFB->np;
   }
                                      /* invert curvature matrix */
   if ( InvMat( psFFB->pdAlpha,
                psFFB->np, (int *) psFFB->plRow, (int *) psFFB->plCol,
                psFFB->pdCovar ) != 0 ) {
      vfill( 0E0, psFFB->pdCovar, 1, psFFB->np * psFFB->np );
      iRC = ERRFIT_INVERTCURVATURE; goto gError;
   }
                                      /* restore old diagonals */
   pdC = psFFB->pdCovar;
   for ( ip1=0; ip1<psFFB->np; ip1++ ) {
      if ( pdD[ip1] != 0E0 ) pdC[ip1] = 0E0;
      pdC += psFFB->np;
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}

/*
   =========================================================================
*/
int fitFctBlkCovarStr( struct SFITFCTBLK *psFFB,
                       char *pcStr,
                       unsigned long lFlag
                     ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkCovarStr)", *pMsg = pDbgMsg;
   int ip1, ip2, nStr = 0;
   double dd;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   if ( psFFB -> pdCovar == NULL ) goto gError;
                                      /* estimated string length */
   nStr = 8 + psFFB->np * 8 +
          psFFB->np * ( 8 + psFFB->np * 8 );

   if ( pcStr == NULL ) {             /* return length */
      iRC = nStr;
   } else {                           /* build string */
      sprintf( pcStr, "   PAR" );
      for ( ip2=0; ip2<psFFB->np-1; ip2++ ) {
         sprintf( pcStr, "%s %6d", pcStr, ip2+1 );
      }
      sprintf( pcStr, "%s\n", pcStr );
      for ( ip1=1; ip1<psFFB->np; ip1++ ) {
         sprintf( pcStr, "%s%6d", pcStr, ip1+1 );
         for ( ip2=0; ip2<ip1; ip2++ ) {
            dd = psFFB->pdCovar[ ip1 * psFFB->np + ip2 ]
               / sqrt( fabs(
                 psFFB->pdCovar[ ip1 * psFFB->np + ip1 ] *
                 psFFB->pdCovar[ ip2 * psFFB->np + ip2 ]
                           ) );
            sprintf( pcStr, "%s %6.3f", pcStr, dd );
         }
         sprintf( pcStr, "%s\n", pcStr );
      }
      iRC = 0;
   }
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}

/*
   =========================================================================
*/
int fitFctBlkCorrCopy( struct SFITFCTBLK *psFFBFr,
                       long lvSubIndex[][2],  /* 0..nn-1 */
                       int nSub,
                       struct SFITFCTBLK *psFFBTo,
                       unsigned long lFlag
                     ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkCorrCopy)", *pMsg = pDbgMsg;
   int ip1Fr, ip2Fr, ip1To, ip2To, is1, is2;
   long lRefTyp, lRefNum, lIndex;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFFBFr == NULL ) goto gError;
   if ( psFFBTo == NULL ) goto gError;
   if ( lvSubIndex == NULL ) goto gError;
   iRC = ERRFIT_BADSUBMATRIXCOUNT;
   if ( nSub <= 0 ) goto gError;
   iRC = ERRFIT_BADPARNUM;
   for ( is1=0; is1<nSub; is1++ ) {
      if ( lvSubIndex[is1][0] < 0 ) goto gError;
      if ( lvSubIndex[is1][1] < 0 ) goto gError;
      if ( lvSubIndex[is1][0] >= psFFBFr -> np ) goto gError;
      if ( lvSubIndex[is1][1] >= psFFBFr -> np ) goto gError;
   }
                                      /* loop over par range to copy */
   ip1To = 0;
   for ( is1=0; is1<nSub; is1++ ) {
                                      /* loop over parameters to copy */
      for ( ip1Fr=lvSubIndex[is1][0]; ip1Fr<=lvSubIndex[is1][1]; ip1Fr++ ) {
         if ( ip1To >= psFFBTo->np ) {
            iRC = ERRFIT_BADPARNUM; goto gError;
         }
         psFFBTo -> pdBeta[ ip1To ] = psFFBFr -> pdBeta[ ip1Fr ];
         psFFBTo -> pdDiag[ ip1To ] = psFFBFr -> pdDiag[ ip1Fr ];
         psFFBTo -> plRefTyp[ ip1To ] = psFFBFr -> plRefTyp[ ip1Fr ];
         psFFBTo -> plRefNum[ ip1To ] = psFFBFr -> plRefNum[ ip1Fr ];
                                      /* loop over par range to copy */
         ip2To = 0;
         for ( is2=0; is2<nSub; is2++ ) {
                                      /* loop over parameters to copy */
            for ( ip2Fr=lvSubIndex[is2][0]; ip2Fr<=lvSubIndex[is2][1]; ip2Fr++ ) {
               if ( ip2To >= psFFBTo->np ) {
                  iRC = ERRFIT_BADPARNUM; goto gError;
               }
               psFFBTo -> pdAlpha[ ip1To * psFFBTo->np + ip2To ] =
               psFFBFr -> pdAlpha[ ip1Fr * psFFBFr->np + ip2Fr ];
               psFFBTo -> pdCovar[ ip1To * psFFBTo->np + ip2To ] =
               psFFBFr -> pdCovar[ ip1Fr * psFFBFr->np + ip2Fr ];
               ip2To++;
            }
         }
         ip1To++;
      }
   }
                                      /* readjust or remove parameter refs */
   for ( ip1To=0; ip1To<psFFBTo->np; ip1To++ ) {
      if ( psFFBTo -> plRefTyp[ ip1To ] > 0 ) {
         lRefTyp = lRefNum = 0;       /* default */
                                      /* look if inside copied interval */
         lIndex = 0;
         for ( is1=0; is1<nSub; is1++ ) {
            if ( ( psFFBTo -> plRefNum[ ip1To ] >= lvSubIndex[is1][0] ) &&
                 ( psFFBTo -> plRefNum[ ip1To ] <= lvSubIndex[is1][1] ) ) {
                                      /* adjust */
               lRefTyp = psFFBTo -> plRefTyp[ ip1To ];
               lRefNum = psFFBTo -> plRefNum[ ip1To ] -
                         lvSubIndex[is1][0] + lIndex;
            }
            lIndex += lvSubIndex[is1][1] - lvSubIndex[is1][0] + 1;
         }
                                      /* insert default or adjusted val.s*/
         psFFBTo -> plRefNum[ ip1To ] = lRefNum;
         psFFBTo -> plRefTyp[ ip1To ] = lRefTyp;
      }
   }
   psFFBTo -> sFS.dLambda = psFFBFr -> sFS.dLambda;
                                      /* loop over par range to copy */
   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}

/*
   =========================================================================
*/
int fitFctBlkParStrByNum( struct SFITFCTBLK *psFFB,
                          long lPar,
                          char *pcStr,
                          unsigned long lFlag
                        ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitFctBlkParStrByNum)", *pMsg = pDbgMsg;
   int nStr = 16 * 10;                /* 16 chars for each of 10 items*/

   struct SFITPAR sFP;
   long lPeak = 0, lOffs = 0, lPNum, lRNum = 0;

   double dSigFac = 1E0;
   char cLim[2][20];
   char *pcPar    = "Param";
   char *pcRefPar = "RefPar";
   char *pcValue  = "Value";
   char *pcFixVar = "Var";
   char *pcEOnOff = "Off";
   char *pcRefTyp;
   char cRefNum[20];

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
   memset( &sFP, 0, sizeof(sFP) );
                                      /* check valid inputs */
   if ( psFFB == NULL ) goto gError;
   iRC = ERRFIT_BADPARNUM;
                                      /* par=0: build header line */
   if ( lPar <= 0 ) {
                                      /* special for peaks */
      switch( lFlag &
              ( FITFCTBLKPAR_AREA | FITFCTBLKPAR_POS | FITFCTBLKPAR_WIDTH )
            ) {
      case FITFCTBLKPAR_AREA:
         pcPar = "Peak#"; pcRefPar = "RefPk#"; pcValue = "Area ";
      break;
      case FITFCTBLKPAR_POS:
         pcPar = "Peak#"; pcRefPar = "RefPk#"; pcValue = "Pos  ";
      break;
      case FITFCTBLKPAR_WIDTH:
         pcPar = "Peak#"; pcRefPar = "RefPk#"; pcValue = "FWHM ";
         if ( lFlag & FITFCTBLKPAR_SIGMA )     pcValue = "Sigma";
      break;
      default: break;
      }

      if ( pcStr == NULL ) {             /* return length */
         iRC = nStr;
      } else {                           /* build string */
         sprintf( pcStr, "%6s %12s %10s %10s %12s %12s %3s %3s %6s %4s",
                         pcPar,
                         pcValue, "Neg Error", "Pos Error",
                         "Lower Limit", "Upper Limit",
                         "F/V", "Err", pcRefPar, "Typ "
                );
         iRC = 0;
      }
      goto gError;
   }
                                      /* fetch parameter values */
   if ( iRC = fitFctBlkParGetByNum( psFFB, lPar, &sFP, lFlag ) ) goto gError;
                                      /* parameter number -> peak number */
   lPNum = lPar;
   if ( lFlag &
           ( FITFCTBLKPAR_AREA | FITFCTBLKPAR_POS | FITFCTBLKPAR_WIDTH )
         ) {
      if ( (iRC = fitFctBlkPar2Peak( psFFB, &lPar, 1, &lPeak, lFlag) ) == 0 ) {
         iRC = fitFctBlkPeak2Par( psFFB, &lPeak, 1, &lOffs, lFlag);
         lOffs = lPar-lOffs;
                                      /* sigma instead FWHM ? */
         if ( lFlag & FITFCTBLKPAR_SIGMA ) {
            if ( lOffs == FITFCT_OFFSPEAKWIDTH ) dSigFac = 1E0 / 2.354820045E0;
         }
         lPNum = lPeak;
      }
   }
                                      /* reference number -> peak number */
   if ( sFP.lRefTyp ) lRNum = sFP.lRefNum + 1; 
   if ( lFlag &
           ( FITFCTBLKPAR_AREA | FITFCTBLKPAR_POS | FITFCTBLKPAR_WIDTH )
         ) {
      if ( sFP.lRefTyp ) {
         if ( (iRC = fitFctBlkPar2Peak( psFFB, &lRNum, 1, &lPeak, lFlag ) ) == 0 ) {
            lRNum = lPeak;
         }
      }
   }

   if ( pcStr == NULL ) {             /* return length */
      iRC = nStr;
   } else {                           /* build string */
      sFP.dVal *= dSigFac;
      sFP.dErr[0] *= dSigFac;
      sFP.dErr[1] *= dSigFac;
      if ( sFP.lFlag & FITPAR_FIX ) pcFixVar = "Fix";
      if ( sFP.lFlag & FITPAR_EON ) pcEOnOff = "On ";

      if ( sFP.dLim[0] > -DBL_MAX ) {
         sprintf( cLim[0], "%12G", sFP.dLim[0] * dSigFac );
      } else {
         sprintf( cLim[0], "%12s", "-INF" );
      }
      if ( sFP.dLim[1] <  DBL_MAX ) {
         sprintf( cLim[1], "%12G", sFP.dLim[1] * dSigFac );
      } else {
         sprintf( cLim[1], "%12s", "+INF" );
      }

      sprintf( cRefNum, "%6d", lRNum );
      switch( sFP.lRefTyp ) {
      case FITPAR_SUM:
         pcRefTyp = "sum ";
      break;
      case FITPAR_DIFF:
         pcRefTyp = "diff";
      break;
      case FITPAR_PROD:
         pcRefTyp = "prod";
      break;
      case FITPAR_QUOT:
         pcRefTyp = "quot";
      break;
      default:
         pcRefTyp = "    ";
         sprintf( cRefNum, "%s", pcRefTyp );
      break;
      }

      sprintf( pcStr, "%6d %12G %10G %10G %s %s %s %s %6s %4s",
                      lPNum,
                      sFP.dVal, sFP.dErr[0], sFP.dErr[1],
                      cLim[0], cLim[1],
                      pcFixVar, pcEOnOff,
                      cRefNum, pcRefTyp
             );
      iRC = 0;
   }
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
