
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "m.h"
#include "g.h"
#include "fit.h"
static char rcsid[] = {"$Id: fitleg.c,v 1.2 1999/08/02 09:21:16 kraemer Exp $"};
/*
   --- Legendre Polynomial functions

       fitLeg()         : basic polynomial
       fitLegEven()     : polynomial, even coeff.
       fitLegOdd()      : polynomial, odd  coeff. + constant
       fitLegNorm()     : basic polynomial, normalized
       fitLegEvenNorm() : polynomial, even coeff., normalized
       fitLegOddNorm () : polynomial, odd  coeff. + constant, normalized

   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
  
   $Log: fitleg.c,v $
   Revision 1.2  1999/08/02 09:21:16  kraemer
   *** empty log message ***

   Revision 1.1.1.1  1997/07/10 18:00:06  kraemer
   revision for cvs


*/
#define DDPAR(ii,jj) ddPar[ (ii) * nX + (jj) ]

/*
   =========================================================================
*/
int fitLeg ( void *pvFF,
             double *dPar,                  /* function parameters */
             double *ddPar,                 /* switch=FITFCT_VAL:
                                              ddPar[nPar][0] = lower limit
                                              ddPar[nPar][1] = upper limit
                                              switch=FITFCT_DERIV:
                                              ddPar[nPar][nX] */
             double *dX, double *dY,
             int nPar, int nX,
             unsigned long lFlag
           ) {
   int iRC = FITUSR_OK;
   struct SFITFCT *psFF = pvFF;
   int nOrder = nPar-1;


   switch( lFlag ) {
                                      /* return # of parameters */
   case FITFCT_NPAR:
      return(nPar);
   break;
                                      /* return start values */
   case FITFCT_VAL:
      memset(  dPar, 0, nPar * sizeof(double) );
   break;
   case FITFCT_FCT:       /* return function values only */
   case FITFCT_DISP:      /* return function values only, for display */
      ddPar = NULL;
   case FITFCT_DERIV:     /* return function values & parameter derivatives  */
      vleg( dX, 1, dPar, nOrder, dY, 1, ddPar, nX );
   break;
   default: break;
   }
   return(iRC);
}
/*
   =========================================================================
*/
int fitLegEven ( void *pvFF,
             double *dPar,                  /* function parameters */
             double *ddPar,                 /* switch=FITFCT_VAL:
                                              ddPar[nPar][0] = lower limit
                                              ddPar[nPar][1] = upper limit
                                              switch=FITFCT_DERIV:
                                              ddPar[nPar][nX] */
             double *dX, double *dY,
             int nPar, int nX,
             unsigned long lFlag
           ) {
   int iRC = FITUSR_OK;
   struct SFITFCT *psFF = pvFF;
   int nOrder = (nPar-1)*2;


   switch( lFlag ) {
                                      /* return # of parameters */
   case FITFCT_NPAR:
      return(nPar);
   break;
                                      /* return start values */
   case FITFCT_VAL:
      memset(  dPar, 0, nPar * sizeof(double) );
   break;
   case FITFCT_FCT:       /* return function values only */
   case FITFCT_DISP:      /* return function values only, for display */
      ddPar = NULL;
   case FITFCT_DERIV:     /* return function values & parameter derivatives  */
      vlegevn( dX, 1, dPar, nOrder, dY, 1, ddPar, nX );
   break;
   default: break;
   }
   return(iRC);
}
/*
   =========================================================================
*/
int fitLegOdd( void *pvFF,
             double *dPar,                  /* function parameters */
             double *ddPar,                 /* switch=FITFCT_VAL:
                                              ddPar[nPar][0] = lower limit
                                              ddPar[nPar][1] = upper limit
                                              switch=FITFCT_DERIV:
                                              ddPar[nPar][nX] */
             double *dX, double *dY,
             int nPar, int nX,
             unsigned long lFlag
           ) {
   int iRC = FITUSR_OK;
   struct SFITFCT *psFF = pvFF;
   int nOrder = (nPar-1)*2;


   switch( lFlag ) {
                                      /* return # of parameters */
   case FITFCT_NPAR:
      return(nPar);
   break;
                                      /* return start values */
   case FITFCT_VAL:
      memset(  dPar, 0, nPar * sizeof(double) );
   break;
   case FITFCT_FCT:       /* return function values only */
   case FITFCT_DISP:      /* return function values only, for display */
      ddPar = NULL;
   case FITFCT_DERIV:     /* return function values & parameter derivatives  */
      vlegodd( dX, 1, dPar, nOrder, dY, 1, ddPar, nX );
   break;
   default: break;
   }
   return(iRC);
}
/*
   =========================================================================
*/
int fitLegNorm ( void *pvFF,
             double *dPar,                  /* function parameters */
             double *ddPar,                 /* switch=FITFCT_VAL:
                                              ddPar[nPar][0] = lower limit
                                              ddPar[nPar][1] = upper limit
                                              switch=FITFCT_DERIV:
                                              ddPar[nPar][nX] */
             double *dX, double *dY,
             int nPar, int nX,
             unsigned long lFlag
           ) {
   int iRC = FITUSR_OK;
   struct SFITFCT *psFF = pvFF;
   int nOrder = nPar-1;


   switch( lFlag ) {
                                      /* return # of parameters */
   case FITFCT_NPAR:
      return(nPar);
   break;
                                      /* return start values */
   case FITFCT_VAL:
      memset(  dPar, 0, nPar * sizeof(double) );
   break;
   case FITFCT_FCT:       /* return function values only */
   case FITFCT_DISP:      /* return function values only, for display */
      ddPar = NULL;
   case FITFCT_DERIV:     /* return function values & parameter derivatives  */
/*
fprintf( stderr, "<D> vlegn(%X,%d,%X,%d,%X,%d,%X,%d)...\n", 
                      dX, 1, dPar, nOrder, dY, 1, ddPar, nX );
*/
      vlegn( dX, 1, dPar, nOrder, dY, 1, ddPar, nX );
/*
fprintf( stderr, "<D> vlegn OK\n" );
*/
   break;
   default: break;
   }
   return(iRC);
}
/*
   =========================================================================
*/
int fitLegEvenNorm ( void *pvFF,
             double *dPar,                  /* function parameters */
             double *ddPar,                 /* switch=FITFCT_VAL:
                                              ddPar[nPar][0] = lower limit
                                              ddPar[nPar][1] = upper limit
                                              switch=FITFCT_DERIV:
                                              ddPar[nPar][nX] */
             double *dX, double *dY,
             int nPar, int nX,
             unsigned long lFlag
           ) {
   int iRC = FITUSR_OK;
   struct SFITFCT *psFF = pvFF;
   int nOrder = (nPar-1)*2;


   switch( lFlag ) {
                                      /* return # of parameters */
   case FITFCT_NPAR:
      return(nPar);
   break;
                                      /* return start values */
   case FITFCT_VAL:
      memset(  dPar, 0, nPar * sizeof(double) );
   break;
   case FITFCT_FCT:       /* return function values only */
   case FITFCT_DISP:      /* return function values only, for display */
      ddPar = NULL;
   case FITFCT_DERIV:     /* return function values & parameter derivatives  */
      vlegevnn( dX, 1, dPar, nOrder, dY, 1, ddPar, nX );
   break;
   default: break;
   }
   return(iRC);
}
/*
   =========================================================================
*/
int fitLegOddNorm( void *pvFF,
             double *dPar,                  /* function parameters */
             double *ddPar,                 /* switch=FITFCT_VAL:
                                              ddPar[nPar][0] = lower limit
                                              ddPar[nPar][1] = upper limit
                                              switch=FITFCT_DERIV:
                                              ddPar[nPar][nX] */
             double *dX, double *dY,
             int nPar, int nX,
             unsigned long lFlag
           ) {
   int iRC = FITUSR_OK;
   struct SFITFCT *psFF = pvFF;
   int nOrder = (nPar-1)*2;


   switch( lFlag ) {
                                      /* return # of parameters */
   case FITFCT_NPAR:
      return(nPar);
   break;
                                      /* return start values */
   case FITFCT_VAL:
      memset(  dPar, 0, nPar * sizeof(double) );
   break;
   case FITFCT_FCT:       /* return function values only */
   case FITFCT_DISP:      /* return function values only, for display */
      ddPar = NULL;
   case FITFCT_DERIV:     /* return function values & parameter derivatives  */
      vlegoddn( dX, 1, dPar, nOrder, dY, 1, ddPar, nX );
   break;
   default: break;
   }
   return(iRC);
}
/*
   =========================================================================
*/
