#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "m.h"
#include "g.h"
#include "fit.h"

static char rcsid[] = {"$Id: fitwdw.c,v 1.2 1997/08/29 12:10:54 kraemer Exp $"};

/*
   --- Functions to handle fit windows

       fitWdwListDup()    : duplicate a list of windows
       fitWdwListInspect(): inspect a list of windows
       fitWdwListFree()   : releases a list of windows
       fitWdwPrompt()     : cursor/terminal prompt a window
       fitWdwAnd()        : overlap 2 windows
       fitWdwInit()       : initialize
       fitWdwInspect()    : inspect output
 
   Copyright  (C) 1987-1997 M.Kraemer, D-64283 Darmstadt, Hoelgesstr. 16,
                             all rights reserved
   $Log: fitwdw.c,v $
   Revision 1.2  1997/08/29 12:10:54  kraemer
   gfxattr rework

   Revision 1.1.1.1  1997/07/10 18:00:07  kraemer
   revision for cvs

*/

int fitWdwListDup( struct List *pslhFr,
                   struct List *pslhTo,
                   unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitWdwListDup)", *pMsg = pDbgMsg;
   struct Node *psln;
   struct SFITWDWNODE *psFWN = NULL;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( pslhTo == NULL ) goto gError;
   NewList( pslhTo );
   if ( pslhFr == NULL ) goto gError;

   for( psln = pslhFr->lh_Head; psln->ln_Succ; psln=psln->ln_Succ ) {
                                      /* allocate new node */
      if ( ( psFWN = calloc( 1, sizeof(*psFWN) ) ) == NULL ) {
         iRC = ERRFIT_SYS; goto gError;
      }
                                      /* copy */
      *psFWN = *((struct SFITWDWNODE *) psln);
                                      /* duplicate node name, if any */
      if ( psln->ln_Name ) {
         if ( (psFWN->ln.ln_Name = strdup( psln->ln_Name )) == NULL ) {
            iRC = ERRFIT_SYS; goto gError;
         }
      }
                                      /* link into new list */
      AddTail( pslhTo, &psFWN->ln ); psFWN = NULL;
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
                                      /* cleanup in case of errors */
   if ( iRC < 0 ) {
      if ( psFWN ) {
         if ( psFWN->ln.ln_Name ) free( psFWN->ln.ln_Name );
         free(psFWN); psFWN = NULL;
      }
      if ( pslhTo ) {
         fitWdwListFree( pslhTo, lFlag );
      }
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitWdwListInspect( struct List *pslh,
                       unsigned long lFlag
                     ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitWdwListInspect)", *pMsg = pDbgMsg;
   struct Node *psln;
   struct SFITWDWNODE *psFWN;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( pslh == NULL ) goto gError;

   fprintf( stderr, "<D> List=%X\n", pslh );
   for( psln = pslh->lh_Head; psln->ln_Succ; psln=psln->ln_Succ ) {
      if ( psln->ln_Name )
         fprintf( stderr, "<D> name=%s\n", psln->ln_Name );
      psFWN = (struct SFITWDWNODE *) psln;

      fitWdwInspect( &psFWN->sFW, lFlag );
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitWdwListFree( struct List *pslh,
                    unsigned long lFlag
                  ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitWdwListFree)", *pMsg = pDbgMsg;
   struct Node *psln;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( pslh == NULL ) goto gError;

   while( pslh -> lh_Head -> ln_Succ ) {
      psln = RemTail( pslh );
      if ( psln -> ln_Name ) free( psln -> ln_Name ); psln -> ln_Name = NULL;
      free( psln ); psln = NULL;
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/

int fitWdwAnd( struct SFITWDW *psFW1,
               struct SFITWDW *psFW2,
               struct SFITWDW *psFW3,
               unsigned long lFlag
             ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitWdwAnd)", *pMsg = pDbgMsg;
   int iDim;
   struct SFITWDW sFW;

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFW1 == NULL ) goto gError;
   if ( psFW2 == NULL ) goto gError;
   if ( psFW3 == NULL ) goto gError;

   if ( psFW1->nDim != psFW2->nDim ) {
      iRC = ERRFIT_WDWDIFFDIM; goto gError;
   }

   fitWdwInit( &sFW, lFlag );
   sFW.nDim = psFW1->nDim;
   for ( iDim=0; iDim<sFW.nDim+1; iDim++ ) {
      sFW.dLim[iDim][0] = max( psFW1->dLim[iDim][0], psFW2->dLim[iDim][0] );
      sFW.dLim[iDim][1] = min( psFW1->dLim[iDim][1], psFW2->dLim[iDim][1] );
   }
   *psFW3 = sFW;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitWdwPrompt( struct SFITWDW *psFW,
                  struct SFITGFX *psFGfx,
                  unsigned long lFlag
                ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitWdwPrompt)", *pMsg = pDbgMsg;
   int iDim;
   short  iOld[2] = {0,0};
   double dWdw[2][2];
   int iCursorType = GCURSOR_XBAR;
   struct SGFXATTR sGA;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFGfx == NULL ) goto gError;
   if ( psFW == NULL ) goto gError;
   fitWdwInit( psFW, lFlag );
                                      /* cursor mode */
   if ( _GWdwIsScreen(psFGfx->pGH) > 0 ) {
      _GGfxAttrInit( &sGA, psFGfx->lFlag );
      _GGfxAttrSet( psFGfx->pGH, &sGA,
                    GGFXATTR_COLOUR | GGFXATTR_FCT | G_DONTPLOT |
                    GGFXATTR_LMODE | GGFXATTR_LWIDTH | psFGfx->lFlag );
      iRC = _GCursor( psFGfx->pGH, iOld, dWdw, iCursorType, psFGfx->lFlag | G_COORDLOG);
      if ( iRC < 0 ) {                /* cursor error ? */
         iRC = ERRFIT_CURSOR;
         goto gError;
      }
                                      /* regular stop */
      if ( iRC > 0 ) goto gError;
                                      /* regular return */
      if ( iRC == 0 ) {
         psFW->dLim[0][0] = min( dWdw[0][0], dWdw[1][0] );
         psFW->dLim[0][1] = max( dWdw[0][0], dWdw[1][0] );
/*
         psFW->dLim[1][0] = min( dWdw[0][1], dWdw[1][1] );
         psFW->dLim[1][1] = max( dWdw[0][1], dWdw[1][1] );
*/
      }
   } else {                           /* terminal mode */
                                      /* old values == default */
      dWdw[0][0] = psFW->dLim[0][0];
      dWdw[0][1] = psFW->dLim[0][1];
      iRC = _pvrl("Window limits (a pair of numbers)", dWdw[0], 2 );
      if ( iRC < 0 ) {                /* cursor error ? */
         iRC = ERRFIT_CURSOR;
         goto gError;
      }
      if ( iRC == 0 ) { iRC = 1; goto gError; } /* blank input stops */
                                      /* return limits */
      psFW->dLim[0][0] = min( dWdw[0][0], dWdw[0][1] );
      psFW->dLim[0][1] = max( dWdw[0][0], dWdw[0][1] );
   }

   psFW -> nDim = 1;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitWdwInit( struct SFITWDW *psFW,
                unsigned long lFlag
              ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitWdwInit)", *pMsg = pDbgMsg;
   int iDim;

if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFW == NULL ) goto gError;

   for ( iDim=0; iDim<MFITWDWDIM+1; iDim++ ) {
      psFW -> dLim[ iDim ][0] = -DBL_MAX;
      psFW -> dLim[ iDim ][1] =  DBL_MAX;
   }
   psFW -> nDim = 0;

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
/*
   =========================================================================
*/
int fitWdwInspect( struct SFITWDW *psFW,
                   unsigned long lFlag
                 ) {
   int iRC = ERRFIT_BADADDR;
   char *pDbgMsg = __FILE__"(fitWdwInspect)", *pMsg = pDbgMsg;
   int iDim;


if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s.....\n", pDbgMsg );
                                      /* check valid inputs */
   if ( psFW == NULL ) goto gError;

   fprintf( stderr, "<D> SFITWDW=%X\n", psFW );
   fprintf( stderr, "<D> nDim=%d\n", psFW->nDim );
   for ( iDim=0; iDim<psFW->nDim+1; iDim++ ) {
      fprintf( stderr, "<D> %d %G %G\n", iDim,
                       psFW->dLim[iDim][0], psFW->dLim[iDim][1]
             );
   }

   iRC = 0;
gError:;
   if ( (iRC < 0) && !(lFlag&FIT_NOMSG) ) {
      fitemsg( pMsg, iRC );
   }
if ( lFlag & FIT_DEBUG ) fprintf( stderr, "<D> %s done\n", pDbgMsg );
   return(iRC);
}
